import numpy as np

try:
    from .utils import *
except (ImportError, ModuleNotFoundError):
    from utils import *


class comparator:  # Class to hold global-type variables

    def __init__(self):
        self.reset()

    def reset(self):
        self.perfect_Starts = 0
        self.perfect_Stops = 0
        self.genome_Seq = ''
        self.genome_Seq_Rev = ''
        self.genome_Size = 0
        self.correct_Frame_Number = 0
        self.extended_Start = 0
        self.extended_Stop = 0
        self.extended_CDS = 0

        self.perfect_Matches = collections.OrderedDict()
        self.matched_ORFs = collections.OrderedDict()
        self.multi_Matched_ORFs = collections.defaultdict(list)
        self.unmatched_ORFs = collections.OrderedDict()
        self.genes_Detected = collections.OrderedDict()
        self.genes_Undetected = collections.OrderedDict()
        self.out_Of_Frame_ORFs = collections.OrderedDict()
        self.partial_Hits = collections.OrderedDict()

        self.start_Difference = []
        self.stop_Difference = []
        self.orf_Lengths = []
        self.gene_Lengths = []

        self.gene_Pos_Olap = []
        self.gene_Neg_Olap = []
        self.orf_Pos_Olap = []
        self.orf_Neg_Olap = []
        self.m_ORF_Pos_Olap = []
        self.m_ORF_Neg_Olap = []

        self.gene_GC = []
        self.orf_GC = []
        self.m_ORF_GC = []

        self.gene_Short = []
        self.orf_Short = []
        self.m_ORF_Short = []

        self.pos_Strand = 0
        self.neg_Strand = 0



comp = comparator()


def is_double_range(range1, range2):
    return len(range1) >= 2 * len(range2)


def nuc_Count(verbose, start, stop, strand):  # Gets correct seq then returns GC
    if stop >= comp.genome_Size:
        if verbose == True:
            print("There is a wrap around gene and I am dealing with it the best I can - Start: " + str(start) + " Stop: " + str(stop))
        extra_stop = stop - comp.genome_Size
        stop = comp.genome_Size
        if strand == '-':
            r_Start = comp.genome_Size - stop
            r_Stop = comp.genome_Size - start
            seq = (comp.genome_Seq_Rev[r_Start:r_Stop + 1])
            extra_seq = (comp.genome_Seq_Rev[-extra_stop-1:])
            seq = extra_seq+seq
        elif strand == '+':
            seq = comp.genome_Seq[start - 1:stop]
            extra_seq = comp.genome_Seq[:extra_stop +1]
            seq = seq+extra_seq
            #seq = (comp.genome_Seq[start - 1:stop])
    else:
        if strand == '-':
            r_Start = comp.genome_Size - stop
            r_Stop = comp.genome_Size - start
            seq = (comp.genome_Seq_Rev[r_Start:r_Stop + 1])
        elif strand == '+':
            seq = (comp.genome_Seq[start - 1:stop])
    c = 0
    a = 0
    g = 0
    t = 0
    n = 0
    for i in seq:
        if "C" in i:
            c += 1
        elif "G" in i:
            g += 1
        elif "A" in i:
            a += 1
        elif "T" in i:
            t += 1
        elif "N" in i:
            n += 1
    gc_content = (g + c) * 100 / (a + t + g + c + n)
    return gc_content


def orf_Unmatched(o_Start, o_Stop, o_Strand, tools):
    if o_Strand == '-':
        r_Start = comp.genome_Size - o_Stop
        r_Stop = comp.genome_Size - o_Start
        Unmatched_ORF = str(o_Start) + ',' + str(o_Stop) + ',' + o_Strand + ',' + comp.genome_Seq_Rev[
                                                                                  r_Start:r_Start + 3] + ',' + comp.genome_Seq_Rev[
                                                                                                               r_Stop - 2:r_Stop + 1] + ',' + tools
        seq = (comp.genome_Seq_Rev[r_Start:r_Stop + 1])
        comp.unmatched_ORFs.update({Unmatched_ORF: seq})
    elif o_Strand == '+':
        Unmatched_ORF = str(o_Start) + ',' + str(o_Stop) + ',' + o_Strand + ',' + comp.genome_Seq[
                                                                                  o_Start - 1:o_Start + 2] + ',' + comp.genome_Seq[
                                                                                                                   o_Stop - 3:o_Stop] + ',' + tools
        seq = (comp.genome_Seq[o_Start - 1:o_Stop])
        comp.unmatched_ORFs.update({Unmatched_ORF: seq})


def genes_Unmatched(g_Start, g_Stop, g_Strand, tools):
    if g_Strand == '-':
        r_Start = comp.genome_Size - g_Stop
        r_Stop = comp.genome_Size - g_Start
        missed_Gene = str(g_Start) + ',' + str(g_Stop) + ',' + g_Strand + ',' + comp.genome_Seq_Rev[
                                                                                r_Start:r_Start + 3] + ',' + comp.genome_Seq_Rev[
                                                                                                             r_Stop - 2:r_Stop + 1] + ',' + tools
        genSeq = (comp.genome_Seq_Rev[r_Start:r_Stop + 1])
        comp.genes_Undetected.update({missed_Gene: genSeq})
    elif g_Strand == '+':
        missed_Gene = str(g_Start) + ',' + str(g_Stop) + ',' + g_Strand + ',' + comp.genome_Seq[
                                                                                g_Start - 1:g_Start + 2] + ',' + comp.genome_Seq[
                                                                                                                 g_Stop - 3:g_Stop] + ',' + tools
        genSeq = (comp.genome_Seq[g_Start - 1:g_Stop])
        comp.genes_Undetected.update({missed_Gene: genSeq})


def perfect_Matched_Genes(g_Start, g_Stop, g_Strand,tools):
    if g_Strand == '-':
        r_Start = comp.genome_Size - g_Stop
        r_Stop = comp.genome_Size - g_Start
        perfect_Matched_Gene = str(g_Start) + ',' + str(g_Stop) + ',' + g_Strand + ',' + comp.genome_Seq_Rev[
                                                                                         r_Start:r_Start + 3] + ',' + comp.genome_Seq_Rev[
                                                                                                                      r_Stop - 2:r_Stop + 1] + ',' + tools
        genSeq = (comp.genome_Seq_Rev[r_Start:r_Stop + 1])
        comp.perfect_Matches.update({perfect_Matched_Gene: genSeq})
    elif g_Strand == '+':
        perfect_Matched_Gene = str(g_Start) + ',' + str(g_Stop) + ',' + g_Strand + ',' + comp.genome_Seq[
                                                                                         g_Start - 1:g_Start + 2] + ',' + comp.genome_Seq[
                                                                                                                          g_Stop - 3:g_Stop] + ',' + tools
        genSeq = (comp.genome_Seq[g_Start - 1:g_Stop])
        comp.perfect_Matches.update({perfect_Matched_Gene: genSeq})


def match_Statistics(o_Start, o_Stop, g_Start, g_Stop, g_Strand):
    comp.correct_Frame_Number += 1
    ############ Calculate prediction precision
    if '+' in g_Strand:
        comp.start_Difference.append(o_Start - g_Start)
        comp.stop_Difference.append(o_Stop - g_Stop)
        if g_Start == o_Start:
            comp.perfect_Starts += 1
        if g_Stop == o_Stop:
            comp.perfect_Stops += 1
        if o_Start < g_Start and o_Stop > g_Stop:
            comp.extended_CDS += 1
        if o_Start < g_Start:
            comp.extended_Start += 1
        if o_Stop > g_Stop:
            comp.extended_Stop += 1
    elif '-' in g_Strand:  # Negative strand genes are reversed
        comp.start_Difference.append(o_Stop - g_Stop)
        comp.stop_Difference.append(o_Start - g_Start)
        if g_Start == o_Start:
            comp.perfect_Stops += 1
        if g_Stop == o_Stop:
            comp.perfect_Starts += 1
        if o_Start < g_Start and o_Stop > g_Stop:
            comp.extended_CDS += 1
        if o_Start < g_Start:
            comp.extended_Stop += 1
        if o_Stop > g_Stop:
            comp.extended_Start += 1


def start_Codon_Count(orfs):
    atg, gtg, ttg, att, ctg, other = 0, 0, 0, 0, 0, 0
    other_Starts = []
    for orf in orfs.values():
        codon = orf[1]
        if codon == 'ATG':
            atg += 1
        elif codon == 'GTG':
            gtg += 1
        elif codon == 'TTG':
            ttg += 1
        elif codon == 'ATT':
            att += 1
        elif codon == 'CTG':
            ctg += 1
        else:
            other += 1
            other_Starts.append(codon)

    total = len(orfs) if orfs is not None else 0

    if total:
        atg_P = format(100 * atg / len(orfs), '.2f')
        gtg_P = format(100 * gtg / len(orfs), '.2f')
        ttg_P = format(100 * ttg / len(orfs), '.2f')
        att_P = format(100 * att / len(orfs), '.2f')
        ctg_P = format(100 * ctg / len(orfs), '.2f')
        other_Start_P = format(100 * other / len(orfs), '.2f')
    else:
        atg_P = ttg_P = gtg_P = ctg_P = att_P = other_Start_P = format(0, '.2f')

    return {
        'ATG': (atg, atg_P),
        'TTG': (ttg, ttg_P),
        'GTG': (gtg, gtg_P),
        'CTG': (ctg, ctg_P),
        'ATT': (att, att_P),
        'Other': (other, other_Start_P),
        'total': total
    }

def stop_Codon_Count(orfs):
    tag, taa, tga, other = 0, 0, 0, 0
    other_Stops = []

    total = len(orfs) if orfs else 0
    if total:
        for orf in orfs.values():
            codon = orf[2]
            if codon == 'TAG':
                tag += 1
            elif codon == 'TAA':
                taa += 1
            elif codon == 'TGA':
                tga += 1
            else:
                other += 1
                other_Stops.append(codon)
        tag_p = format(100 * tag / len(orfs), '.2f')
        taa_p = format(100 * taa / len(orfs), '.2f')
        tga_p = format(100 * tga / len(orfs), '.2f')
        other_Stop_P = format(100 * other / len(orfs), '.2f')
    else:
        tag_p = taa_p = tga_p = other_Stop_P = format(0, '.2f')

    return tag_p, taa_p, tga_p, other_Stop_P, other_Stops


def candidate_ORF_Selection(gene_Set,
                            candidate_ORFs):  # Select ORF from candidates which is most similar to partially detected gene
    current_Coverage = 0
    candidate_ORF_Difference = 0
    pos = ''
    orf_Details = []
    for c_Pos, c_ORF_Details in candidate_ORFs.items():
        o_Start = int(c_Pos.split(',')[0])
        o_Stop = int(c_Pos.split(',')[1])
        # Below is not a long term fix
        coverage = c_ORF_Details[-1]
        orf_Set = set(range(o_Start, o_Stop + 1))
        if coverage > current_Coverage:
            current_Coverage = coverage
            # Return set of elements outside the two sets/DNA ranges
            candidate_ORF_Difference = orf_Set.symmetric_difference(gene_Set)
            pos = c_Pos
            orf_Details = c_ORF_Details
        elif coverage == current_Coverage:
            current_ORF_Difference = orf_Set.symmetric_difference(
                gene_Set)  # Pick least different ORF set from the Gene Set
            if len(current_ORF_Difference) > len(candidate_ORF_Difference):
                pos = c_Pos
                orf_Details = c_ORF_Details
        #else:
            #("Match filtered out")
    return pos, orf_Details


def partial_Hit_Calc(g_Start, g_Stop, g_Strand, o_Start, o_Stop, tools):
    if g_Strand == '-':
        r_G_Start = comp.genome_Size - g_Stop
        r_G_Stop = comp.genome_Size - g_Start
        r_O_Start = comp.genome_Size - o_Stop
        r_O_Stop = comp.genome_Size - o_Start
        partial = "Gene:" + str(g_Start) + '_' + str(g_Stop) + '_' + g_Strand + '_' + comp.genome_Seq_Rev[
                                                                                      r_G_Start:r_G_Start + 3] + '_' + comp.genome_Seq_Rev[
                                                                                                                       r_G_Stop - 2:r_G_Stop + 1] + ';Predicted_CDS:' + str(
            o_Start) + '_' + str(o_Stop) + '_' + g_Strand + '_' + comp.genome_Seq_Rev[
                                                                  r_O_Start:r_O_Start + 3] + '_' + comp.genome_Seq_Rev[
                                                                                                   r_O_Stop - 2:r_O_Stop + 1] + ';' + tools
        genSeq = (comp.genome_Seq_Rev[r_G_Start:r_G_Stop + 1])
        orfSeq = (comp.genome_Seq_Rev[r_O_Start:r_O_Stop + 1])
        comp.partial_Hits.update({partial: [genSeq, orfSeq]})
    elif g_Strand == '+':
        partial = "Gene:" + str(g_Start) + '_' + str(g_Stop) + '_' + g_Strand + '_' + comp.genome_Seq[
                                                                                      g_Start - 1:g_Start + 2] + '_' + comp.genome_Seq[
                                                                                                                       g_Stop - 3:g_Stop] + ';Predicted_CDS:' + str(
            o_Start) + '_' + str(o_Stop) + '_' + g_Strand + '_' + comp.genome_Seq[
                                                                  o_Start - 1:o_Start + 2] + '_' + comp.genome_Seq[
                                                                                                   o_Stop - 3:o_Stop] + ';' + tools
        genSeq = (comp.genome_Seq[g_Start - 1:g_Stop])
        orfSeq = (comp.genome_Seq[o_Start - 1:o_Stop])
        comp.partial_Hits.update({partial: [genSeq, orfSeq]})


def tool_comparison(all_orfs, dna_regions, verbose):
    results = collections.OrderedDict()  # Store results for each DNA region
    for dna_region in dna_regions:  # Loop through each DNA region
        # reset comparator class variables
        comp.reset()

        ref_genes_list = dna_regions[dna_region][2]
        ref_genes = collections.OrderedDict()

        if not ref_genes_list:
            results[dna_region] = {}
            continue

        for d in ref_genes_list:
            ref_genes.update(d)
        comp.genome_Seq = dna_regions[dna_region][0]
        comp.genome_Seq_Rev = revCompIterative(dna_regions[dna_region][0])
        comp.genome_Size = len(dna_regions[dna_region][0])

        current_orfs = all_orfs[dna_region]
        # sort the ORFs by start position

        better_pos_orfs_items = [[(int(pos.split(',')[0]), int(pos.split(',')[1])), orf_Details] for pos, orf_Details in current_orfs.items()] #TODO: turn pos into tuple instead of string everywhere

        if not current_orfs or not better_pos_orfs_items:
            results[dna_region] = {}
            continue

        for gene_num, gene_details in ref_genes.items():  # Loop through each gene to compare against predicted ORFs
            g_Start = int(gene_details[0])
            g_Stop = int(gene_details[1])
            g_Strand = gene_details[2]
            g_pos = str(g_Start) + ',' + str(g_Stop)
            gene_Set = set(range(g_Start,
                                 g_Stop + 1))  # Used to check Overlap of ORFs and pick best match - slow but confirms best match
            overlapping_ORFs = collections.OrderedDict()
            perfect_Match = False
            out_Frame = False
            for pos, orf_Details in better_pos_orfs_items:  # Check if perfect match, if not check if match covers at least 75% of gene - Loop through ALL ORFs - SLOW
                o_Start,o_Stop = pos
                o_Strand = orf_Details[0]
                #orf_Set = set(range(o_Start, o_Stop + 1)) Removed for optimisation
                if o_Stop <= g_Start or o_Start >= g_Stop:  # Not caught up yet
                    continue
                elif o_Start == g_Start and o_Stop == g_Stop:  # If perfect match, break and skip the rest of the ORFs
                    perfect_Match = True
                    break
                elif is_double_range(range(o_Start, o_Stop), range(g_Start,g_Stop)):  # If ORF is double or more than the length of the gene, we do not count as found.
                    continue
                elif g_Start <= o_Start < g_Stop or g_Start < o_Stop < g_Stop:  # If ORF Start or Stop is between gene Start or Stop
                    #overlap = len(gene_Set.intersection(orf_Set)) # Replaced for optimisation
                    overlap = max(min(o_Stop, g_Stop) - max(o_Start, g_Start), -1) + 1
                    coverage = 100 * float(overlap) / float(len(gene_Set))
                    orf_Details.append(coverage)
                    if abs(o_Stop - g_Stop) % 3 == 0 and o_Strand == g_Strand and coverage >= MIN_COVERAGE:  # Only continue if ORF covers at least 75% of the gene and is in frame
                        overlapping_ORFs.update({f'{o_Start},{o_Stop}': orf_Details})
                    elif coverage >= MIN_COVERAGE:  # Not in frame / on same strand
                        comp.out_Of_Frame_ORFs.update({f'{o_Start},{o_Stop}': orf_Details})
                        out_Frame = True
                elif o_Start <= g_Start and o_Stop >= g_Stop:  # If ORF extends one or both ends of the gene
                    #overlap = len(gene_Set.intersection(orf_Set)) # Replaced for optimisation
                    overlap = max(min(o_Stop, g_Stop) - max(o_Start, g_Start), -1) + 1
                    coverage = 100 * float(overlap) / float(len(gene_Set))
                    orf_Details.append(coverage)
                    if abs(o_Stop - g_Stop) % 3 == 0 and o_Strand == g_Strand and coverage >= MIN_COVERAGE:  # Only continue if ORF covers at least 75% of the gene and is in frame
                        overlapping_ORFs.update({f'{o_Start},{o_Stop}': orf_Details})
                    elif coverage >= MIN_COVERAGE:
                        comp.out_Of_Frame_ORFs.update({f'{o_Start},{o_Stop}': orf_Details})
                        out_Frame = True
                else:
                    if verbose == True:
                        print("Unexpected Error Finding Predicted CDSs")  # Should not happen
            # Now Check that we select the best ORF
            ### Multi_Match_ORFs Should contain All genes found by a specific ORF
            if perfect_Match == True:  # Check if the ORF is a perfect match to the Gene
                m_ORF_Details = orf_Details[:]
                m_ORF_Details.append(g_pos)
                if g_pos in comp.matched_ORFs.keys():
                    previously_Covered_Gene = comp.matched_ORFs[g_pos][-1]
                    comp.multi_Matched_ORFs[g_pos] += [g_pos.replace(',', '-'), previously_Covered_Gene.replace(',',
                                                                                                                '-'), orf_Details[4]]  # ORF is same as gene so can use g_pos
                comp.matched_ORFs.update({g_pos: m_ORF_Details})
                comp.genes_Detected.update({str(gene_details): g_pos})
                match_Statistics(o_Start, o_Stop, g_Start, g_Stop, g_Strand)
                perfect_Matched_Genes(g_Start, g_Stop, g_Strand,orf_Details[4])
                #if verbose == True:
                #    print('Perfect Match')
            elif perfect_Match == False and len(
                    overlapping_ORFs) == 1:  # If we do not have a perfect match but 1 ORF which has passed the filtering
                orf_Pos = list(overlapping_ORFs.keys())[0]
                o_Start = int(orf_Pos.split(',')[0])
                o_Stop = int(orf_Pos.split(',')[1])
                orf_Details = overlapping_ORFs[orf_Pos]
                m_ORF_Details = orf_Details[:]
                m_ORF_Details.append(g_pos)
                if orf_Pos in comp.matched_ORFs.keys():
                    try:
                        previously_Covered_Gene = comp.matched_ORFs[g_pos][-1]
                    except KeyError:
                        last_key = [*comp.matched_ORFs.keys()][-1]
                        previously_Covered_Gene = comp.matched_ORFs[last_key][-1]
                    comp.multi_Matched_ORFs[orf_Pos] += [g_pos.replace(',', '-'), previously_Covered_Gene.replace(',',
                                                                                                       '-'), orf_Details[4]]  # ORF collects multiple gene pos'
                comp.matched_ORFs.update({orf_Pos: m_ORF_Details})
                comp.genes_Detected.update({str(gene_details): orf_Pos})
                match_Statistics(o_Start, o_Stop, g_Start, g_Stop, g_Strand)
                #if verbose == True:
                #    print('Partial Match')
                partial_Hit_Calc(g_Start, g_Stop, g_Strand, o_Start, o_Stop, orf_Details[4])
            elif perfect_Match == False and len(
                    overlapping_ORFs) >= 1:  # If we have more than 1 potential ORF match, we check to see which is the 'best' hit
                orf_Pos, orf_Details = candidate_ORF_Selection(gene_Set, overlapping_ORFs)  # Return best match
                o_Start = int(orf_Pos.split(',')[0])
                o_Stop = int(orf_Pos.split(',')[1])
                m_ORF_Details = orf_Details[:]
                m_ORF_Details.append(g_pos)
                if orf_Pos in comp.matched_ORFs.keys():
                    try:
                        previously_Covered_Gene = comp.matched_ORFs[g_pos][-1]
                    except KeyError:
                        last_key = [*comp.matched_ORFs.keys()][-1]
                        previously_Covered_Gene = comp.matched_ORFs[last_key][-1]
                    comp.multi_Matched_ORFs[orf_Pos] += [g_pos.replace(',', '-'), previously_Covered_Gene.replace(',',
                                                                                                                  '-'), orf_Details[4]]  # ORF collects multiple gene pos'
                comp.matched_ORFs.update({orf_Pos: m_ORF_Details})
                comp.genes_Detected.update({str(gene_details): orf_Pos})
                match_Statistics(o_Start, o_Stop, g_Start, g_Stop, g_Strand)
                if verbose == True:
                    print('There was more than 1 potential Match - Best Chosen')
                partial_Hit_Calc(g_Start, g_Stop, g_Strand, o_Start, o_Stop, orf_Details[4])
            elif out_Frame:  # Keep record of ORFs which overlap a gene but in the wrong frame
                if verbose == True:
                    print("Out of Frame Predicted CDS")
                genes_Unmatched(g_Start, g_Stop, g_Strand, orf_Details[4])  #
            else:
                genes_Unmatched(g_Start, g_Stop, g_Strand, orf_Details[4])  # No hit
                #if verbose == True:
                #    print("No Hit")
        for orf_Key in comp.matched_ORFs:  # Remove ORFs from out of frame if ORF was correctly matched to another Gene
            if orf_Key in comp.out_Of_Frame_ORFs:
                del comp.out_Of_Frame_ORFs[orf_Key]
        ######################################## ORF Lengths and Precision
        start_Difference = [x for x in comp.start_Difference if x != 0]  # Remove 0s (Perfect hits)
        stop_Difference = [x for x in comp.stop_Difference if x != 0]
        if len(start_Difference) >= 1:
            median_Start_Difference = np.median(start_Difference)
        else:
            median_Start_Difference = 'N/A'
        if len(stop_Difference) >= 1:
            median_Stop_Difference = np.median(stop_Difference)
        else:
            median_Stop_Difference = 'N/A'

        # Get Start and Stop Codon Usage
        atg_P, gtg_P, ttg_P, att_P, ctg_P, other_Start_P, other_Starts = start_Codon_Count(current_orfs)
        tag_P, taa_P, tga_P, other_Stop_P, other_Stops = stop_Codon_Count(current_orfs)
        # Count nucleotides found from ALL ORFs
        gene_Nuc_Array = np.zeros((comp.genome_Size), dtype=bool)
        orf_Nuc_Array = np.zeros((comp.genome_Size), dtype=bool)
        matched_ORF_Nuc_Array = np.zeros((comp.genome_Size), dtype=bool)

        prev_Gene_Stop = 0
        prev_Gene_Overlapped = False
        for gene_Num, gene_details in ref_genes.items():  # Loop through each gene to compare against predicted ORFs
            g_Start = int(gene_details[0])
            g_Stop = int(gene_details[1])
            g_Strand = gene_details[2]
            gene_Length = (g_Stop - g_Start) +1
            if gene_Length == 0: print(g_Start, g_Stop, "!!!!!!!!!!!!!!!!!!!!!!!!")
            comp.gene_Lengths.append(gene_Length)
            gene_Nuc_Array[g_Start - 1:g_Stop] = True  # Changing all between the two positions to 1's
            comp.gene_GC.append(nuc_Count(verbose, g_Start, g_Stop, g_Strand))
            if gene_Length <= SHORT_ORF_LENGTH:  # .utils
                comp.gene_Short.append(gene_Length)
            ### Calculate overlapping Genes -
            if prev_Gene_Stop > g_Start:
                if '+' in g_Strand:
                    comp.gene_Pos_Olap.append(prev_Gene_Stop - g_Start)
                elif '-' in g_Strand:
                    comp.gene_Neg_Olap.append(prev_Gene_Stop - g_Start)
                prev_Gene_Overlapped = True
            elif prev_Gene_Stop < g_Start:
                if prev_Gene_Overlapped == True:
                    if '+' in g_Strand:
                        comp.gene_Pos_Olap.append(0)
                    elif '-' in g_Strand:
                        comp.gene_Neg_Olap.append(0)
                prev_Gene_Overlapped = False
            prev_Gene_Stop = g_Stop
        if prev_Gene_Overlapped == True:  # If last has a prev overlap, count it
            if '+' in g_Strand:
                comp.gene_Pos_Olap.append(0)
            elif '-' in g_Strand:
                comp.gene_Neg_Olap.append(0)
        #### avoid ValueError
        if comp.gene_Lengths:
            min_Gene_Length = min(comp.gene_Lengths)
            max_Gene_Length = max(comp.gene_Lengths)
            median_Gene_Length = np.median(comp.gene_Lengths)
        else:
            min_Gene_Length = max_Gene_Length = min_Length_Difference = 0
        prev_ORF_Stop = 0
        prev_ORF_Overlapped = False
        for o_Positions, orf_Details in current_orfs.items():
            o_Start = int(o_Positions.split(',')[0])
            o_Stop = int(o_Positions.split(',')[1])
            o_Strand = orf_Details[0]
            # Stats just for Unmatched ORFs
            if o_Positions not in list(comp.matched_ORFs.keys()):
                orf_Unmatched(o_Start, o_Stop, o_Strand, orf_Details[4])
            # Get ORF Strand metrics:
            if o_Strand == "+":  # Get number of Positive and Negative strand ORFs
                comp.pos_Strand += 1
            elif o_Strand == "-":
                comp.neg_Strand += 1
            orf_Length = (o_Stop - o_Start) +1
            comp.orf_Lengths.append(orf_Length)
            orf_Nuc_Array[o_Start - 1:o_Stop] = True  # Changing all between the two positions to 1's
            comp.orf_GC.append(nuc_Count(verbose, o_Start, o_Stop, o_Strand))
            if orf_Length <= SHORT_ORF_LENGTH:  # .utils
                comp.orf_Short.append(orf_Length)
            ### Calculate overlapping ORFs -
            if prev_ORF_Stop > o_Start:
                if '+' in o_Strand:
                    comp.orf_Pos_Olap.append(prev_ORF_Stop - o_Start)
                elif '-' in o_Strand:
                    comp.orf_Neg_Olap.append(prev_ORF_Stop - o_Start)
                prev_ORF_Overlapped = True
            elif prev_ORF_Stop < o_Start:
                if prev_ORF_Overlapped == True:
                    if '+' in o_Strand:
                        comp.orf_Pos_Olap.append(0)
                    elif '-' in o_Strand:
                        comp.orf_Neg_Olap.append(0)
                prev_ORF_Overlapped = False
            prev_ORF_Stop = o_Stop
            if prev_ORF_Overlapped == True:  # If last has a prev overlap, count it
                if '+' in o_Strand:
                    comp.orf_Pos_Olap.append(0)
                elif '-' in o_Strand:
                    comp.orf_Neg_Olap.append(0)

        # Nucleotide Coverage calculated from ORFs matching a gene only
        matched_Prev_ORF_Stop = 0
        matched_Prev_ORF_Overlapped = False
        for mo_Positions, m_ORF_Details in comp.matched_ORFs.items():
            mo_Start = int(mo_Positions.split(',')[0])
            mo_Stop = int(mo_Positions.split(',')[1])
            mo_Strand = m_ORF_Details[0]
            mo_Length = (mo_Stop - mo_Start)
            matched_ORF_Nuc_Array[mo_Start - 1:mo_Stop] = True  # This is the complete matched orf not the matched orf bits

            comp.m_ORF_GC.append(nuc_Count(verbose, mo_Start, mo_Stop, mo_Strand))
            if mo_Length <= SHORT_ORF_LENGTH:  # .utils
                comp.m_ORF_Short.append(mo_Length)
            ### Calculate overlapping Matched ORFs -
            if matched_Prev_ORF_Stop > mo_Start:
                if '+' in mo_Strand:
                    comp.m_ORF_Pos_Olap.append(matched_Prev_ORF_Stop - mo_Start)
                elif '-' in mo_Strand:
                    comp.m_ORF_Neg_Olap.append(matched_Prev_ORF_Stop - mo_Start)
                matched_Prev_ORF_Overlapped = True
            elif matched_Prev_ORF_Stop < mo_Start:
                if matched_Prev_ORF_Overlapped == True:
                    if '+' in mo_Strand:
                        comp.m_ORF_Pos_Olap.append(0)
                    elif '-' in mo_Strand:
                        comp.m_ORF_Neg_Olap.append(0)
                matched_Prev_ORF_Overlapped = False
            matched_Prev_ORF_Stop = mo_Stop
        if matched_Prev_ORF_Overlapped == True:  # If last has a prev overlap, count it
            if '+' in mo_Strand:
                comp.m_ORF_Pos_Olap.append(0)
            elif '-' in mo_Strand:
                comp.m_ORF_Neg_Olap.append(0)
        ####
        gene_Coverage_Genome = format(100 * np.sum(gene_Nuc_Array) / comp.genome_Size, '.2f')
        orf_Coverage_Genome = format(100 * np.sum(orf_Nuc_Array) / comp.genome_Size, '.2f')
        matched_ORF_Coverage_Genome = format(100 * np.sum(matched_ORF_Nuc_Array) / comp.genome_Size,
                                             '.2f')  # This gets the nts which are in matched ORFs - Check below
        # matched_ORF_Nuc_AND_Gene = np.logical_and(matched_ORF_Nuc_Array,gene_Nuc_Array) + [0 for i in range(len(gene_Nuc_Array))] # This gets the nts which are in both matched ORFs and detected genes
        # matched_ORF_Coverage_Genome = format(100 * np.count_nonzero(matched_ORF_Nuc_AND_Gene) / comp.genome_Size,'.2f')

        # gene and orf nucleotide Intersection
        gene_ORF_Nuc_Intersection = np.sum(gene_Nuc_Array & orf_Nuc_Array)
        # not gene but orf nucleotides
        not_Gene_Nuc_Array = np.logical_not(gene_Nuc_Array)
        not_Gene_Nuc_And_ORF_Count = np.sum(not_Gene_Nuc_Array & orf_Nuc_Array)
        # not orf nucleotides but gene
        not_ORF_Nuc_Array = np.logical_not(orf_Nuc_Array)
        not_ORF_Nuc_And_Gene_Count = np.sum(not_ORF_Nuc_Array & gene_Nuc_Array)
        # not gene or orf nucleotides
        not_Gene_Nuc_Not_ORF_Nuc_Count = np.sum(not_Gene_Nuc_Array & not_ORF_Nuc_Array)
        # Nucleotide 'accuracy' - Normalised by number of nucelotides annotated by a gene
        NT_TP = format(gene_ORF_Nuc_Intersection / np.sum(gene_Nuc_Array), '.2f')
        NT_FP = format(not_Gene_Nuc_And_ORF_Count / np.sum(not_Gene_Nuc_Array), '.2f')
        NT_FN = format(not_ORF_Nuc_And_Gene_Count / np.sum(gene_Nuc_Array), '.2f')
        NT_TN = format(not_Gene_Nuc_Not_ORF_Nuc_Count / np.sum(not_Gene_Nuc_Array), '.2f')
        NT_Precision = format(gene_ORF_Nuc_Intersection / (gene_ORF_Nuc_Intersection + not_Gene_Nuc_And_ORF_Count), '.2f')
        NT_Recall = format(gene_ORF_Nuc_Intersection / (gene_ORF_Nuc_Intersection + not_ORF_Nuc_And_Gene_Count), '.2f')
        NT_False_Discovery_Rate = format(
            not_Gene_Nuc_And_ORF_Count / (not_Gene_Nuc_And_ORF_Count + gene_ORF_Nuc_Intersection), '.2f')
        ################################# Precision and Recall of whole ORFs and Genes
        TP = format(len(comp.genes_Detected) / len(ref_genes), '.2f')
        FP = format(len(comp.unmatched_ORFs) / len(ref_genes), '.2f')
        FN = format(len(comp.genes_Undetected) / len(ref_genes), '.2f')
        #################################################### Need a better way to handle 'no hits/ORFs'
        try:
            precision = format(float(TP) / (float(TP) + float(FP)), '.2f')
        except ZeroDivisionError:
            precision = format(0.00, '.2f')
        try:
            recall = format(float(TP) / (float(TP) + float(FN)), '.2f')
        except ZeroDivisionError:
            recall = format(0.00, '.2f')
        try:
            false_Discovery_Rate = format(float(FP) / (float(FP) + float(TP)), '.2f')
        except ZeroDivisionError:
            false_Discovery_Rate = 'N/A'
        min_ORF_Length = min(comp.orf_Lengths)
        max_ORF_Length = max(comp.orf_Lengths)
        median_ORF_Length = np.median(comp.orf_Lengths)

        ##########################################################################
        # Metrics - There are numerous cases where certain metric calculations may return a ZeroDivError.
        ORFs_Difference = format(100 * (len(current_orfs) - len(ref_genes)) / len(ref_genes), '.2f')  # Difference off +/-
        genes_Detected_Percentage = format(100 * (len(comp.genes_Detected) / len(ref_genes)), '.2f')
        matched_ORF_Percentage = format(100 * (len(comp.matched_ORFs) / len(current_orfs)), '.2f')
        all_ORF_Olap = (comp.orf_Pos_Olap + comp.orf_Neg_Olap)  # Combine pos and neg strand overlaps
        matched_ORF_Olap = (comp.m_ORF_Pos_Olap + comp.m_ORF_Neg_Olap)
        all_Gene_Olap = (comp.gene_Pos_Olap + comp.gene_Neg_Olap)

        if all_ORF_Olap:  # If no overlapping ORFs
            try:
                overlap_Difference = format(100 * (len(all_ORF_Olap) - len(all_Gene_Olap)) / len(all_Gene_Olap), '.2f')
                matched_Overlap_Difference = format(100 * (len(matched_ORF_Olap) - len(all_Gene_Olap)) / len(all_Gene_Olap),
                                                    '.2f')
            except ZeroDivisionError:
                overlap_Difference = 'N/A'
                matched_Overlap_Difference = 'N/A'
            num_All_ORF_Olap = len(all_ORF_Olap)
            if matched_ORF_Olap:
                max_Matched_ORF_Olap = max(matched_ORF_Olap)
                matched_Median_ORF_Overlap = format(np.median(matched_ORF_Olap), '.2f')
            else:
                max_Matched_ORF_Olap = 'N/A'
                matched_Median_ORF_Overlap = 'N/A'
            max_All_ORF_Olap = max(all_ORF_Olap)
            median_ORF_Overlap = format(np.median(all_ORF_Olap), '.2f')
        else:
            overlap_Difference = 'N/A'
            matched_Overlap_Difference = 'N/A'
            num_All_ORF_Olap = 0
            max_Matched_ORF_Olap = 'N/A'
            max_All_ORF_Olap = 'N/A'
            median_ORF_Overlap = 'N/A'
            matched_Median_ORF_Overlap = 'N/A'
        if len(matched_ORF_Olap) == 0:  # -100.00 is not informative
            matched_Overlap_Difference = 'N/A'

        # Need to NA everything

        if comp.orf_Short and comp.gene_Short:  # IF Short-ORFs/Genes
            short_ORF_Difference = format(100 * (len(comp.orf_Short) - len(comp.gene_Short)) / len(comp.gene_Short), '.2f')
            matched_Short_ORF_Difference = format(
                100 * (len(comp.m_ORF_Short) - len(comp.gene_Short)) / len(comp.gene_Short), '.2f')
            num_ORF_Short = len(comp.orf_Short)
            num_Matched_ORF_Short = len(comp.m_ORF_Short)
        elif comp.orf_Short:  # If only Short-ORFs
            num_ORF_Short = len(comp.orf_Short)
            num_Matched_ORF_Short = 'N/A'
            short_ORF_Difference = (num_ORF_Short * 100)
            matched_Short_ORF_Difference = 'N/A'
        else:  # If only Short-Genes and Undetected StORFs
            comp.gene_Short
            short_ORF_Difference = 'N/A'
            matched_Short_ORF_Difference = 'N/A'
            num_ORF_Short = 0
            num_Matched_ORF_Short = 'N/A'
        if num_Matched_ORF_Short == 0:  # -100.00 is not informative
            matched_Short_ORF_Difference = 'N/A'

        median_Length_Difference = format(100 * (median_ORF_Length - median_Gene_Length) / median_Gene_Length, '.2f')
        min_Length_Difference = format(100 * (min_ORF_Length - min_Gene_Length) / min_Gene_Length, '.2f')
        max_Length_Difference = format(100 * (max_ORF_Length - max_Gene_Length) / max_Gene_Length, '.2f')
        pos_Strand_Percentage = format(comp.pos_Strand / len(current_orfs), '.2f')
        neg_Strand_Percentage = format(comp.neg_Strand / len(current_orfs), '.2f')
        median_ORF_GC = np.median(comp.orf_GC)
        matched_Median_ORF_GC = np.median(comp.m_ORF_GC)
        median_Gene_GC = np.median(comp.gene_GC)
        median_GC_Difference = format(100 * (float(median_ORF_GC) - float(median_Gene_GC)) / float(median_Gene_GC), '.2f')
        matched_Median_GC_Difference = format(
            100 * (float(matched_Median_ORF_GC) - float(median_Gene_GC)) / float(median_Gene_GC), '.2f')

        if comp.matched_ORFs:  # No ORFs detected a gene
            extended_CDS_Percentage = format(100 * comp.extended_CDS / len(comp.matched_ORFs), '.2f')
            extended_Start_Percentage = format(100 * comp.extended_Start / len(comp.matched_ORFs), '.2f')
            extended_Stop_Percentage = format(100 * comp.extended_Stop / len(comp.matched_ORFs), '.2f')
            perfect_Matches_Percentage = format(100 * len(comp.perfect_Matches) / len(comp.matched_ORFs), '.2f')
            perfect_Starts_Percentage = format(100 * comp.perfect_Starts / len(comp.matched_ORFs), '.2f')
            perfect_Stops_Percentage = format(100 * comp.perfect_Stops / len(comp.matched_ORFs), '.2f')
        else:
            # correct_Frame_Percentage = 0
            extended_CDS_Percentage = format(0.00, '.2f')
            extended_Start_Percentage = format(0.00, '.2f')
            extended_Stop_Percentage = format(0.00, '.2f')
            perfect_Matches_Percentage = format(0.00, '.2f')
            perfect_Starts_Percentage = format(0.00, '.2f')
            perfect_Stops_Percentage = format(0.00, '.2f')
        ################### Missed Genes  Metrics:
        if comp.genes_Undetected:
            mg_Starts = []
            mg_Stops = []
            mg_Lengths = []
            mg_Strands = []
            for mg, seq in comp.genes_Undetected.items():
                mg = mg.split(',')
                mg_Starts.append(mg[3])
                mg_Stops.append(mg[4])
                mg_Strands.append(mg[2])
                mg_Lengths.append(int(mg[1]) - int(mg[0]))

            mg_ATG = 100 * mg_Starts.count('ATG') / len(comp.genes_Undetected)
            mg_GTG = 100 * mg_Starts.count('GTG') / len(comp.genes_Undetected)
            mg_TTG = 100 * mg_Starts.count('TTG') / len(comp.genes_Undetected)
            mg_ATT = 100 * mg_Starts.count('ATT') / len(comp.genes_Undetected)
            mg_CTG = 100 * mg_Starts.count('CTG') / len(comp.genes_Undetected)
            mg_O_Start = 100 - (mg_ATG + mg_GTG + mg_TTG + mg_ATT + mg_CTG)
            mg_TGA = 100 * mg_Stops.count('TGA') / len(comp.genes_Undetected)
            mg_TAA = 100 * mg_Stops.count('TAA') / len(comp.genes_Undetected)
            mg_TAG = 100 * mg_Stops.count('TAG') / len(comp.genes_Undetected)
            mg_O_Stop = 100 - (mg_TGA + mg_TAA + mg_TAG)
            median_mg_Len = np.median(mg_Lengths)
            mg_Pos = mg_Strands.count('+')
            mg_Neg = mg_Strands.count('-')
            undetected_Gene_Metrics = (
            format(mg_ATG, '.2f'), format(mg_GTG, '.2f'), format(mg_TTG, '.2f'), format(mg_ATT, '.2f'),
            format(mg_CTG, '.2f'), format(mg_O_Start, '.2f'), format(mg_TGA, '.2f'), format(mg_TAA, '.2f'),
            format(mg_TAG, '.2f'), format(mg_O_Stop, '.2f'), format(median_mg_Len, '.2f'), mg_Pos, mg_Neg)
        else:
            undetected_Gene_Metrics = ''
        #################### Unmathced ORF Metrics:
        if comp.unmatched_ORFs:
            uo_Starts = []
            uo_Stops = []
            uo_Lengths = []
            uo_Strands = []
            for uo, seq in comp.unmatched_ORFs.items():
                uo = uo.split(',')
                uo_Starts.append(uo[3])
                uo_Stops.append(uo[4])
                uo_Strands.append(uo[2])
                uo_Lengths.append(int(uo[1]) - int(uo[0]))
            uo_ATG = 100 * uo_Starts.count('ATG') / len(comp.unmatched_ORFs)
            uo_GTG = 100 * uo_Starts.count('GTG') / len(comp.unmatched_ORFs)
            uo_TTG = 100 * uo_Starts.count('TTG') / len(comp.unmatched_ORFs)
            uo_ATT = 100 * uo_Starts.count('ATT') / len(comp.unmatched_ORFs)
            uo_CTG = 100 * uo_Starts.count('CTG') / len(comp.unmatched_ORFs)
            uo_O_Start = 100 - (uo_ATG + uo_GTG + uo_TTG + uo_ATT + uo_CTG)
            uo_TGA = 100 * uo_Stops.count('TGA') / len(comp.unmatched_ORFs)
            uo_TAA = 100 * uo_Stops.count('TAA') / len(comp.unmatched_ORFs)
            uo_TAG = 100 * uo_Stops.count('TAG') / len(comp.unmatched_ORFs)
            uo_O_Stop = 100 - (uo_TGA + uo_TAA + uo_TAG)
            # uo_O_Stop = 100 * uo_O_Stop  / len(comp.unmatched_ORFs) ########WHY?
            median_uo_Len = np.median(uo_Lengths)
            uo_Pos = uo_Strands.count('+')
            uo_Neg = uo_Strands.count('-')
            unmatched_ORF_Metrics = (
            format(uo_ATG, '.2f'), format(uo_GTG, '.2f'), format(uo_TTG, '.2f'), format(uo_ATT, '.2f'),
            format(uo_CTG, '.2f'), format(uo_O_Start, '.2f'), format(uo_TGA, '.2f'), format(uo_TAA, '.2f'),
            format(uo_TAG, '.2f'), format(uo_O_Stop, '.2f'), format(median_uo_Len, '.2f'), uo_Pos, uo_Neg)
        else:
            unmatched_ORF_Metrics = ''
        #################################
        # Rep_Metrics - This is the final report of metrics
        rep_Metrics = collections.OrderedDict(
            {'Percentage_of_Genes_Detected': genes_Detected_Percentage,
             'genes_Undetected': comp.genes_Undetected,
             'undetected_Gene_Metrics': undetected_Gene_Metrics,
             'gene_Coverage_Genome': gene_Coverage_Genome,
             'Percentage_of_ORFs_that_Detected_a_Gene': matched_ORF_Percentage,
             'Percent_Difference_of_All_ORFs': ORFs_Difference,
             'Median_Length_Difference': median_Length_Difference,
             'Percentage_of_Perfect_Matches': perfect_Matches_Percentage,
             'Median_Start_Difference_of_Matched_ORFs': median_Start_Difference,
             'Median_Stop_Difference_of_Matched_ORFs': median_Stop_Difference,
             'Percentage_Difference_of_Matched_Overlapping_CDSs': matched_Overlap_Difference,
             'Percent_Difference_of_Short-Matched-ORFs': matched_Short_ORF_Difference,
             'Precision': precision,
             'Recall': recall,
             'False_Discovery_Rate': false_Discovery_Rate})
        # Pred Metrics - This is the final report of metrics
        pred_metrics = collections.OrderedDict(
            {'Number_of_ORFs': len(current_orfs), 'Percent_Difference_of_All_ORFs': ORFs_Difference,
             'perfect_Matches': comp.perfect_Matches,
             'unmatched_ORFs': comp.unmatched_ORFs,
             'unmatched_ORF_Metrics': unmatched_ORF_Metrics,
             'orf_Coverage_Genome': orf_Coverage_Genome,
             'matched_ORF_Coverage_Genome': matched_ORF_Coverage_Genome,
             'multi_Matched_ORFs': comp.multi_Matched_ORFs,
             'partial_Hits': comp.partial_Hits,
             'Number_of_ORFs_that_Detected_a_Gene': len(comp.matched_ORFs),
             'Percentage_of_ORFs_that_Detected_a_Gene': matched_ORF_Percentage,
             'Number_of_Genes_Detected': len(comp.genes_Detected),
             'Percentage_of_Genes_Detected': genes_Detected_Percentage, 'Median_Length_of_All_ORFs': median_ORF_Length,
             'Median_Length_Difference': median_Length_Difference,
             'Minimum_Length_of_All_ORFs': min_ORF_Length, 'Minimum_Length_Difference': min_Length_Difference,
             'Maximum_Length_of_All_ORFs': max_ORF_Length, 'Maximum_Length_Difference': max_Length_Difference,
             'Median_GC_content_of_All_ORFs': format(median_ORF_GC, '.2f'),
             'Percent_Difference_of_All_ORFs_Median_GC': median_GC_Difference,
             'Median_GC_content_of_Matched_ORFs': format(matched_Median_ORF_GC, '.2f'),
             'Percent_Difference_of_Matched_ORF_GC': matched_Median_GC_Difference,
             'Number_of_ORFs_which_Overlap_Another_ORF': num_All_ORF_Olap,
             'Percent_Difference_of_Overlapping_ORFs': overlap_Difference,
             'Maximum_ORF_Overlap': max_All_ORF_Olap, 'Median_ORF_Overlap': median_ORF_Overlap,
             'Number_of_Matched_ORFs_Overlapping_Another_ORF': len(matched_ORF_Olap),
             'Percentage_Difference_of_Matched_Overlapping_CDSs': matched_Overlap_Difference,
             'Maximum_Matched_ORF_Overlap': max_Matched_ORF_Olap, 'Median_Matched_ORF_Overlap': matched_Median_ORF_Overlap,
             'Number_of_Short-ORFs': num_ORF_Short, 'Percent_Difference_of_Short-ORFs': short_ORF_Difference,
             'Number_of_Short-Matched-ORFs': num_Matched_ORF_Short,
             'Percent_Difference_of_Short-Matched-ORFs': matched_Short_ORF_Difference,
             'Number_of_Perfect_Matches': len(comp.perfect_Matches),
             'Percentage_of_Perfect_Matches': perfect_Matches_Percentage,
             'Number_of_Perfect_Starts': comp.perfect_Starts, 'Percentage_of_Perfect_Starts': perfect_Starts_Percentage,
             'Number_of_Perfect_Stops': comp.perfect_Stops, 'Percentage_of_Perfect_Stops': perfect_Stops_Percentage,
             'Number_of_Out_of_Frame_ORFs': len(comp.out_Of_Frame_ORFs),
             'Number_of_Matched_ORFs_Extending_a_Coding_Region': comp.extended_CDS,
             'Percentage_of_Matched_ORFs_Extending_a_Coding_Region': extended_CDS_Percentage,
             'Number_of_Matched_ORFs_Extending_Start_Region': comp.extended_Start,
             'Percentage_of_Matched_ORFs_Extending_Start_Region': extended_Start_Percentage,
             'Number_of_Matched_ORFs_Extending_Stop_Region': comp.extended_Stop,
             'Percentage_of_Matched_ORFs_Extending_Stop_Region': extended_Stop_Percentage,
             'Number_of_All_ORFs_on_Positive_Strand': comp.pos_Strand,
             'Percentage_of_All_ORFs_on_Positive_Strand': pos_Strand_Percentage,
             'Number_of_All_ORFs_on_Negative_Strand': comp.neg_Strand,
             'Percentage_of_All_ORFs_on_Negative_Strand': neg_Strand_Percentage,
             'Median_Start_Difference_of_Matched_ORFs': median_Start_Difference,
             'Median_Stop_Difference_of_Matched_ORFs': median_Stop_Difference, 'ATG_Start_Percentage': atg_P,
             'GTG_Start_Percentage': gtg_P, 'TTG_Start_Percentage': ttg_P,
             'ATT_Start_Percentage': att_P, 'CTG_Start_Percentage': ctg_P, 'Other_Start_Codon_Percentage': other_Start_P,
             'TAG_Stop_Percentage': tag_P, 'TAA_Stop_Percentage': taa_P,
             'TGA_Stop_Percentage': tga_P, 'Other_Stop_Codon_Percentage': other_Stop_P, 'True_Positive': TP,
             'False_Positive': FP, 'False_Negative': FN, 'Precision': precision,
             'Recall': recall, 'False_Discovery_Rate': false_Discovery_Rate, 'Nucleotide_True_Positive': NT_TP,
             'Nucleotide_False_Positive': NT_FP, 'Nucleotide_True_Negative': NT_TN,
             'Nucleotide_False_Negative': NT_FN, 'Nucleotide_Precision': NT_Precision, 'Nucleotide_Recall': NT_Recall,
             'Nucleotide_False_Discovery_Rate': NT_False_Discovery_Rate,
             'ORF_Nucleotide_Coverage_of_Genome': orf_Coverage_Genome,
             'Matched_ORF_Nucleotide_Coverage_of_Genome': matched_ORF_Coverage_Genome})
        result = collections.OrderedDict()
        result.update({
            'rep_metrics': rep_Metrics,
            'pred_metrics': pred_metrics,
        })

        # To account for unbalanced data
        for m_key, m_value in result.items():
            if m_value == 'nan':
                result[m_key] = 'N/A'

        results[dna_region] = result


        print("Finished calculating metrics for: ", dna_region)


    return results  # Return the results dictionary containing all metrics and details



import collections
import logging



def _make_feature(seqid, source, type_, start, end, score, strand, phase, attributes):
    attrs = []
    for k, v in attributes.items():
        attrs.append(f"{k}={v}")
    return f"{seqid}\t{source}\t{type_}\t{start}\t{end}\t{score}\t{strand}\t{phase}\t{';'.join(attrs)}\n"


def parse_blast_tab6(path, genome_seq, gene_ident=None):
    results = collections.OrderedDict()
    count = 0
    with open(path, 'r') as fh:
        for i, line in enumerate(fh, 1):
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            parts = line.split('\t')
            if len(parts) < 12:
                logging.warning(f"Line {i}: unexpected BLAST line with {len(parts)} columns")
                continue
            qseqid, sseqid, pident, length, mismatch, gapopen, qstart, qend, sstart, send, evalue, bitscore = parts[:12]
            try:
                sstart = int(sstart)
                send = int(send)
            except ValueError:
                logging.warning(f"Line {i}: non-integer coordinates in BLAST sstart/send")
                continue
            start = min(sstart, send)
            end = max(sstart, send)
            strand = '+' if sstart <= send else '-'
            attrs = {
                'ID': f'blast_hit{count}',
                'Target': f'{qseqid} {qstart} {qend}',
                'pident': pident,
                'length': length,
                'evalue': evalue,
                'bitscore': bitscore
            }
            results[f"{start},{end}"] = [strand, '.', 'similarity', attrs]
            count += 1
    return results


def parse_abricate(path, genome_seq, gene_ident=None):
    results = collections.OrderedDict()
    count = 0
    with (open(path, 'r') as fh):
        header = None
        for i, line in enumerate(fh, 1):
            line = line.rstrip('\n')
            if not line:
                continue
            if line.startswith('#'):
                header = line.split('\t')
                continue
            if header is None:
                # skip any pre-header content until a header line is encountered
                continue
            parts = line.split('\t')
            if header and len(parts) == len(header):
                row = dict(zip(header, parts))

                try:
                    start = int(row.get('START', '0'))
                    end = int(row.get('END', '0'))
                    strand = row.get('STRAND')
                except ValueError:
                    logging.warning(f"Line {i}: invalid START/END in Abricate line")
                    continue
                seqid = row.get('SEQUENCE')
                gene = row.get('GENE')
                accession =  row.get('ACCESSION') or 'unknown'
                db = row.get('DATABASE')  or 'unknown'
                identity = row.get('%IDENTITY')
                coverage = row.get('%COVERAGE')
                product = row.get('PRODUCT') or 'unkown'
                resistance = row.get('RESISTANCE') or 'unknown'

                attrs = {
                    'seqid': seqid,
                    'start': start,
                    'end': end,
                    'strand': strand,
                    'gene': gene,
                    'accession': accession,
                    'database': db,
                    'identity': identity,
                    'coverage': coverage,
                    'product': product,
                    'resistance': resistance
                }
                results[f"{start},{end}"] = attrs
                count += 1
            else:
                logging.warning(f"Line {i}: unexpected number of columns in Abricate line")
                continue
    return results


def parse_genemark(path, genome_seq, gene_ident=None):
    results = collections.OrderedDict()
    count = 0
    with open(path, 'r') as fh:
        for i, line in enumerate(fh, 1):
            line = line.strip()
            if not line:
                continue
            parts = line.split()
            if len(parts) < 3:
                continue
            try:
                start = int(parts[0])
                stop = int(parts[1])
            except ValueError:
                continue
            strand_tok = parts[2]
            if 'complement' in strand_tok:
                strand = '-'
            else:
                strand = '+'
            attrs = {'ID': f'genemark_hit{count}', 'tool': 'GeneMark'}
            results[f"{start},{stop}"] = [strand, '.', 'CDS', attrs]
            count += 1
    return results


def TabToGFF(input_file, genome_seq, gene_ident='CDS', fmt='blast'):
    # Should be cleaned up to use consistent format names
    fmt = fmt.lower()
    if fmt in ('blast', 'blast_tab6', 'tab6'):
        return parse_blast_tab6(input_file, genome_seq, gene_ident)
    if fmt in ('abricate', 'abricate_tsv', 'abricate_format'):
        return parse_abricate(input_file, genome_seq, gene_ident)
    if fmt in ('genemark', 'gene_mark'):
        return parse_genemark(input_file, genome_seq, gene_ident)
    raise ValueError(f"Unknown format: {fmt}")



from importlib import import_module
import argparse
import sys, os
import gzip, csv
import logging
from datetime import datetime


try:
    from .Comparator import tool_comparison
    from .utils import *
except (ImportError, ModuleNotFoundError):
    from Comparator import tool_comparison
    from utils import *


##########################

# Consolidate printing and logging into a single block
def _pct(n, total):
    try:
        return format(100 * n / total, '.2f') + '%'
    except Exception:
        return 'N/A'

##########################

def comparator(options):


    try:
        try:  # Detect whether fasta/gff files are .gz or text and read accordingly
            fasta_in = gzip.open(options.genome_dna, 'rt')
            dna_regions = fasta_load(fasta_in)
        except:
            fasta_in = open(options.genome_dna, 'r', encoding='unicode_escape')
            dna_regions = fasta_load(fasta_in)
        try:
            gff_in = gzip.open(options.reference_annotation, 'rt')
            dna_regions = gff_load(options, gff_in, dna_regions)
        except:
            gff_in = open(options.reference_annotation, 'r', encoding='unicode_escape')
            dna_regions = gff_load(options, gff_in, dna_regions)
    except AttributeError:
        sys.exit("Attribute Error:\nStORF'ed GFF probably already exists - Must be deleted before running (-overwrite)")
    except FileNotFoundError:
        split_path = options.gff.split(os.sep)
        sys.exit("Directory '" + split_path[-2] + "' missing fna/gff files")
        ###############################################
    total_ref_genes = sum(
        len(v[2]) if isinstance(v[2], (list, tuple, set, dict, str)) else 1 for v in dna_regions.values())

    #############################################
    try:
        tool_ = import_module('Tools.' + options.tool + '.' + options.tool, package='my_current_pkg')
    except ModuleNotFoundError:
        try:
            tool_ = import_module('ORForise.Tools.' + options.tool + '.' + options.tool, package='my_current_pkg')
        except ModuleNotFoundError:
            sys.exit("Tool not available - Did you get the name right?")
    tool_ = getattr(tool_, options.tool)
    all_orfs = tool_(options.tool_prediction, dna_regions)
    results = tool_comparison(all_orfs, dna_regions, options.verbose)
    ############## Printing to std-out and optional csv file
    # Ensure the output directory exists
    os.makedirs(options.outdir, exist_ok=True)
    # Use outname as a directory, basename for files is output-outname
    base_out = os.path.join(options.outdir, f"{os.path.basename(options.outname)}")

    # Prepare to collect summary stats for all contigs
    contig_summaries = []

    if options.outdir:
        # Ensure the output directory exists
        os.makedirs(options.outdir, exist_ok=True)
        # Use outname as a directory, basename for files is output-outname
        base_out = os.path.join(options.outdir, f"{os.path.basename(options.outname)}")
        with open(f"{base_out}_summary.txt", 'w', encoding='utf-8') as out_file:
            out_file.write('Genome Used: ' + str(options.genome_dna.split('/')[-1]) + '\n')
            if options.reference_tool:
                out_file.write('Reference Tool Used: ' + str(options.reference_tool) + '\n')
            else:
                out_file.write('Reference Used: ' + str(options.reference_annotation.split('/')[-1]) + '\n')
            out_file.write('Tool Compared: ' + str(options.tool) + '\n')
            out_file.write('Total Number of Reference Genes: ' + str(total_ref_genes) + '\n')
            out_file.write('Number of Contigs: ' + str(len(dna_regions)) + '\n')
            out_file.write(
                'Contig\tGenes\tORFs\tPerfect_Matches\tPartial_Matches\tMissed_Genes\tUnmatched_ORFs\tMulti_Matched_ORFs\n')

    for dna_region, result in results.items():
        if result:
            num_current_genes = len(dna_regions[dna_region][2])
            num_orfs = result['pred_metrics']['Number_of_ORFs']
            num_perfect = result['pred_metrics']['Number_of_Perfect_Matches']
            num_partial = len(result['pred_metrics']['partial_Hits'])
            num_missed = len(result['rep_metrics']['genes_Undetected'])
            num_unmatched = len(result['pred_metrics']['unmatched_ORFs'])
            num_multi = len(result['pred_metrics']['multi_Matched_ORFs'])
            # Collect summary for this contig
            contig_summaries.append([dna_region, num_current_genes, num_orfs, num_perfect, num_partial, num_missed, num_unmatched, num_multi])
            num_current_genes = len(dna_regions[dna_region][2])
            genome_name = options.reference_annotation.split('/')[-1].split('.')[0]
            rep_metric_description, rep_metrics = get_rep_metrics(result)
            all_metric_description, all_metrics = get_all_metrics(result)

             # Safely extract metric values
            num_orfs = result.get('pred_metrics', {}).get('Number_of_ORFs') if isinstance(result, dict) else 'N/A'
            perfect = result.get('pred_metrics', {}).get('Number_of_Perfect_Matches') if isinstance(result, dict) else 0
            partial = len(result.get('pred_metrics', {}).get('partial_Hits', [])) if isinstance(result, dict) else 'N/A'
            missed = len(result.get('rep_metrics', {}).get('genes_Undetected', [])) if isinstance(result, dict) else 'N/A'
            unmatched = len(result.get('pred_metrics', {}).get('unmatched_ORFs', [])) if isinstance(result, dict) else 'N/A'
            multi = len(result.get('pred_metrics', {}).get('multi_Matched_ORFs', [])) if isinstance(result, dict) else 'N/A'

            lines = [
                f"These are the results for: {dna_region}",
                f"Current Contig: {dna_region}",
                f"Number of Genes: {num_current_genes}",
                f"Number of ORFs: {num_orfs}",
                f"Perfect Matches: {perfect} [{num_current_genes}] - {_pct(perfect, num_current_genes) if isinstance(num_current_genes, (int, float)) else 'N/A'}",
                f"Partial Matches: {partial} [{num_current_genes}] - {_pct(partial, num_current_genes) if isinstance(num_current_genes, (int, float)) else 'N/A'}",
                f"Missed Genes: {missed} [{num_current_genes}] - {_pct(missed, num_current_genes) if isinstance(num_current_genes, (int, float)) else 'N/A'}",
                f"Unmatched ORFs: {unmatched} [{num_current_genes}] - {_pct(unmatched, num_current_genes) if isinstance(num_current_genes, (int, float)) else 'N/A'}",
                f"Multi-matched ORFs: {multi} [{num_current_genes}] - {_pct(multi, num_current_genes) if isinstance(num_current_genes, (int, float)) else 'N/A'}"
            ]

            full_msg = '\n'.join(lines) + '\n'
            if options.verbose:
                print(full_msg)
            options.output_logger.info(full_msg)

            # Prepare output directory and file names for each contig
            contig_save = dna_region.replace('/', '_').replace('\\', '_')
            contig_dir = os.path.join(options.outdir, contig_save)
            os.makedirs(contig_dir, exist_ok=True)
            summary_file = os.path.join(contig_dir, "summary.txt")
            csv_file = os.path.join(contig_dir, "metrics.csv")
            perfect_fasta = os.path.join(contig_dir, "perfect_matches.fasta")
            partial_fasta = os.path.join(contig_dir, "partial_matches.fasta")
            missed_fasta = os.path.join(contig_dir, "missed_genes.fasta")
            unmatched_fasta = os.path.join(contig_dir, "unmatched_orfs.fasta")
            multi_fasta = os.path.join(contig_dir, "multi_matched_orfs.fasta")

            # Write summary to text file
            with open(summary_file, 'w', encoding='utf-8') as sf:
                sf.write('Current Contig: ' + str(dna_region) + '\n')
                sf.write('Number of Genes: ' + str(num_current_genes) + '\n')
                sf.write('Number of ORFs: ' + str(result['pred_metrics']['Number_of_ORFs']) + '\n')
                sf.write('Perfect Matches: ' + str(result['pred_metrics']['Number_of_Perfect_Matches']) + ' [' + str(
                    num_current_genes) + '] - ' + format(
                    100 * result['pred_metrics']['Number_of_Perfect_Matches'] / num_current_genes, '.2f') + '%\n')
                sf.write('Partial Matches: ' + str(len(result['pred_metrics']['partial_Hits'])) + ' [' + str(
                    num_current_genes) + '] - ' + format(
                    100 * len(result['pred_metrics']['partial_Hits']) / num_current_genes, '.2f') + '%\n')
                sf.write('Missed Genes: ' + str(len(result['rep_metrics']['genes_Undetected'])) + ' [' + str(
                    num_current_genes) + '] - ' + format(
                    100 * len(result['rep_metrics']['genes_Undetected']) / num_current_genes, '.2f') + '%\n')
                sf.write('Unmatched ORFs: ' + str(len(result['pred_metrics']['unmatched_ORFs'])) + ' [' + str(
                    num_current_genes) + '] - ' + format(
                    100 * len(result['pred_metrics']['unmatched_ORFs']) / num_current_genes, '.2f') + '%\n')
                sf.write('Multi-matched ORFs: ' + str(len(result['pred_metrics']['multi_Matched_ORFs'])) + ' [' + str(
                    num_current_genes) + '] - ' + format(
                    100 * len(result['pred_metrics']['multi_Matched_ORFs']) / num_current_genes, '.2f') + '%\n')


            # Write metrics to CSV
                tool_out = csv.writer(out_file, quoting=csv.QUOTE_NONE, escapechar=" ")
                tool_out = csv.writer(out_file, quoting=csv.QUOTE_NONE, escapechar=" ")  # type: ignore[arg-type]
                tool_out.writerow(['Representative_Metrics:'])
                tool_out.writerow(rep_metric_description.split(','))
                tool_out.writerow([*rep_metrics])
                tool_out.writerow(['Prediction_Metrics:'])
                tool_out.writerow(all_metric_description.split(','))
                tool_out.writerow([*all_metrics])
                tool_out.writerow(['Reference_CDS_Gene_Coverage_of_Genome'])
                tool_out.writerow([''.join(map(str, result['rep_metrics']['gene_Coverage_Genome']))])
                tool_out.writerow(['Predicted_CDS_Coverage_of_Genome'])
                tool_out.writerow([''.join(map(str, result['pred_metrics']['orf_Coverage_Genome']))])
                tool_out.writerow(['Matched_Predicted_CDS_Coverage_of_Genome'])
                tool_out.writerow([''.join(map(str, result['pred_metrics']['matched_ORF_Coverage_Genome']))])

            # Write perfect matches to FASTA
            with open(perfect_fasta, 'w', encoding='utf-8') as f:
                for key, value in result['pred_metrics'].get('perfect_Matches', {}).items():
                    key_parts = key.split(',')
                    id = f">{genome_name}_{key_parts[0]}_{key_parts[1]}_{key_parts[2]}_{key_parts[5]}"
                    f.write(f"{id}\n{value}\n")

            # Write partial matches to FASTA
            with open(partial_fasta, 'w', encoding='utf-8') as f:
                for key, value in result['pred_metrics'].get('partial_Hits', {}).items():
                    key_parts = key.split(';')
                    gene_Seq = value[0]
                    orf_Seq = value[1]
                    f.write(f">{key_parts[0]}_gene\n{gene_Seq}\n>{key_parts[1]}_orf\n{orf_Seq}\n")

            # Write missed genes to FASTA
            with open(missed_fasta, 'w', encoding='utf-8') as f:
                for key, value in result['rep_metrics'].get('genes_Undetected', {}).items():
                    key_parts = key.split(',')
                    id = f">{genome_name}_{key_parts[0]}_{key_parts[1]}_{key_parts[2]}"
                    f.write(f"{id}\n{value}\n")

            # Write unmatched ORFs to FASTA
            with open(unmatched_fasta, 'w', encoding='utf-8') as f:
                for key, value in result['pred_metrics'].get('unmatched_ORFs', {}).items():
                    key_parts = key.split(',')
                    id = f">{options.tool}_{key_parts[0]}_{key_parts[1]}_{key_parts[2]}"
                    f.write(f"{id}\n{value}\n")

            # Write multi-matched ORFs to FASTA
            with open(multi_fasta, 'w', encoding='utf-8') as f:
                for key, value in result['pred_metrics'].get('multi_Matched_ORFs', {}).items():
                    key_parts = key.split(',')
                    multi = f">Predicted_CDS:{key_parts[0]}-{key_parts[1]}_Genes:{'|'.join(value)}"
                    f.write(f"{multi}\n")
        else:
            if options.verbose:
                print(f"No results to process for dna region - " + str(dna_region))
            options.output_logger.info(f"No results to process for dna region - " + str(dna_region))


    # After all contigs, append the summary table to the main summary file
    if options.outdir and contig_summaries:
        with open(f"{base_out}_summary.txt", 'a', encoding='utf-8') as out_file:
            for row in contig_summaries:
                out_file.write('\t'.join(map(str, row)) + '\n')
            # Optionally, add overall totals
            total_genes = sum(row[1] for row in contig_summaries)
            total_orfs = sum(row[2] for row in contig_summaries)
            total_perfect = sum(row[3] for row in contig_summaries)
            total_partial = sum(row[4] for row in contig_summaries)
            total_missed = sum(row[5] for row in contig_summaries)
            total_unmatched = sum(row[6] for row in contig_summaries)
            total_multi = sum(row[7] for row in contig_summaries)
            out_file.write('\nOverall Summary:\n')
            out_file.write(f'Number of Genes: {total_genes}\n')
            out_file.write(f'Number of ORFs: {total_orfs}\n')
            out_file.write(f'Perfect Matches: {total_perfect} [{total_genes}] - {100 * total_perfect / total_genes:.2f}%\n')
            out_file.write(f'Partial Matches: {total_partial} [{total_genes}] - {100 * total_partial / total_genes:.2f}%\n')
            out_file.write(f'Missed Genes: {total_missed} [{total_genes}] - {100 * total_missed / total_genes:.2f}%\n')
            out_file.write(f'Unmatched ORFs: {total_unmatched} [{total_genes}] - {100 * total_unmatched / total_genes:.2f}%\n')
            out_file.write(f'Multi-matched ORFs: {total_multi} [{total_genes}] - {100 * total_multi / total_genes:.2f}%\n')

            lines = [
                f"Combined metrics for all contigs:",
                f"Number of Genes: {total_genes}",
                f"Number of ORFs: {total_orfs}",
                f"Perfect Matches: {total_perfect} [{total_genes}] - {100 * total_perfect / total_genes:.2f}%",
                f"Partial Matches: {total_partial} [{total_genes}] - {100 * total_partial / total_genes:.2f}%",
                f"Missed Genes: {total_missed} [{total_genes}] - {100 * total_missed / total_genes:.2f}%",
                f"Unmatched ORFs: {total_unmatched} [{total_genes}] - {100 * total_unmatched / total_genes:.2f}%",
                f"Multi-matched ORFs: {total_multi} [{total_genes}] - {100 * total_multi / total_genes:.2f}%"
            ]

            full_msg = '\n'.join(lines) + '\n'
            if options.verbose:
                print(full_msg)
            options.output_logger.info(full_msg)


def main():
    parser = argparse.ArgumentParser(description='ORForise ' + ORForise_Version + ': Annotatione-Compare Run Parameters.')
    parser._action_groups.pop()

    required = parser.add_argument_group('Required Arguments')
    required.add_argument('-dna', dest='genome_dna', required=True, help='Genome DNA file (.fa) which both annotations '
                                                                    'are based on')
    required.add_argument('-ref', dest='reference_annotation', required=True,
                        help='Which reference annotation file to use as reference?')
    required.add_argument('-t', dest='tool', required=True, help='Which tool to analyse?')
    required.add_argument('-tp', dest='tool_prediction', required=True,
                        help='Tool genome prediction file (.gff) - Different Tool Parameters'
                             ' are compared individually via separate files')

    optional = parser.add_argument_group('Optional Arguments')
    optional.add_argument('-gene_ident', action='store', dest='gene_ident', default='CDS',
                          help='What features to consider as genes? - Default: CDS - '
                               'Provide comma separated list of features to consider as genes (e.g. CDS,exon)')
    optional.add_argument('-rt', dest='reference_tool', required=False,
                          help='What type of Annotation to compare to? -- Leave blank for Ensembl reference'
                               '- Provide tool name to compare output from two tools')

    output = parser.add_argument_group('Output')
    output.add_argument('-o', dest='outdir', required=True,
                        help='Define directory where detailed output should be places')
    output.add_argument('-n', dest='outname', required=False,
                        help='Define output filename(s) prefix - If not provided, filename of reference '
                             'annotation file will be used- <outname>_<contig_id>_ORF_Comparison.csv')

    misc = parser.add_argument_group('Misc')
    misc.add_argument('-v', dest='verbose', default='False', type=eval, choices=[True, False],
                      help='Default - False: Print out runtime status')
    options = parser.parse_args()

    options.outname = options.outname if options.outname else options.reference_annotation.split('/')[-1].split('.')[0]

    # Initialise loggers once and store on options
    if not getattr(options, 'logger_initialized', False):
        os.makedirs(options.outdir, exist_ok=True)
        output_log = os.path.join(options.outdir, f"ORForise_{options.outname}_{datetime.now().strftime('%Y%m%d_%H%M%S')}.log")
        logger = logging.getLogger('ORForise.output')
        logger.setLevel(logging.INFO)
        fh_out = logging.FileHandler(output_log, encoding='utf-8')
        fh_out.setFormatter(logging.Formatter('%(message)s'))
        logger.addHandler(fh_out)

        options.output_logger = logger
        options.logger_initialized = True


    comparator(options)

if __name__ == "__main__":
    try:
        try:
            main()
        except Exception:
            logging.exception('Unhandled exception in main')
    finally:
        print(CLOSING)


from importlib import import_module
import argparse
from collections import OrderedDict, defaultdict, Counter
import gzip
from datetime import date
import sys

try:
    from .utils import *
except (ImportError, ModuleNotFoundError):
    from utils import *


########################################
def gff_writer(options, combined_ORFs_by_contig, output_file, reference_annotation, additional_annotation):
    write_out = open(output_file, 'w')
    write_out.write("##gff-version\t3\n#\tGFF-Adder\n#\tRun Date:" + str(date.today()) + '\n')
    write_out.write("##Genome DNA File:" + (options.genome_DNA if hasattr(options, 'genome_DNA') else '') + '\n')
    write_out.write("##Original File: " + reference_annotation + "\n##Additional File: " + additional_annotation + '\n')
    # meta counts
    Ref_Only = 0
    Ref_Combined = defaultdict(int)
    Non_Ref_Combined = defaultdict(int)

    # New counters: per-tool totals and per-contig per-tool breakdown
    per_tool_total = Counter()                       # counts for additional-only entries
    per_contig_per_tool = defaultdict(Counter)      # contig -> Counter(tool -> count)
    ref_per_tool_total = Counter()                  # counts for reference entries matched by tools
    ref_per_contig_per_tool = defaultdict(Counter)  # contig -> Counter(tool -> count)

    # Iterate contigs in deterministic order
    for contig in combined_ORFs_by_contig:
        combined_ORFs = combined_ORFs_by_contig[contig]
        # ref_gene_set for this contig: use keys from ref portion (we can detect by data[1]=='ref')
        ref_gene_set = [k for k, v in combined_ORFs.items() if len(v) > 1 and v[1] == 'ref']

        for pos, data in combined_ORFs.items():
            pos_ = pos.split(',')
            # pos may be like 'start,stop' or 'contig,start,stop' but here we expect 'start,stop'
            if len(pos_) >= 2:
                start = pos_[0]
                stop = pos_[-1]
            else:
                # fallback: skip malformed
                continue
            strand = data[0]

            # Build additional_annotation_info from the combined entry's additional list if present.
            # Normalise entries and prefer the info portion after any 'ToolName:info' prefix.
            additional_annotation_info = ''
            additional_items = []
            if len(data) > 4 and data[4]:
                for add in data[4]:
                    s = str(add).strip()
                    if not s:
                        continue
                    if ':' in s:
                        # split tool:info -> take info part
                        _, info_part = s.split(':', 1)
                        info_part = info_part.strip()
                    else:
                        info_part = s
                    if info_part:
                        additional_items.append(info_part)
                # deduplicate while preserving order
                seen = set()
                deduped = []
                for it in additional_items:
                    if it not in seen:
                        seen.add(it)
                        deduped.append(it)
                if deduped:
                    additional_annotation_info = ';'.join(deduped)
            elif len(data) > 3 and data[3]:
                additional_annotation_info = str(data[3]).strip()

            # tools list from options (maybe empty)
            tools = options.additional_tool.split(',') if getattr(options, 'additional_tool', None) else []

            # Determine matched_tools_list reliably:
            # prefer extracting tool names from data[4] entries, otherwise fallback to scanning values
            matched_tools_list = []
            try:
                if len(data) > 4 and data[4]:
                    for add in data[4]:
                        # expected format: "ToolName:info" or "info"
                        if isinstance(add, str) and ':' in add:
                            t = add.split(':', 1)[0].strip()
                            if t:
                                matched_tools_list.append(t)
                        else:
                            # try to detect one of the known tool names in the string
                            for tool in tools:
                                if tool and tool in str(add):
                                    matched_tools_list.append(tool)
                else:
                    # fallback: scan the whole data structure for tool names (previous behaviour)
                    for tool in tools:
                        if tool and any(tool in str(x) for x in data):
                            matched_tools_list.append(tool)
            except Exception:
                # keep matched_tools_list empty if anything unexpected happens
                matched_tools_list = []

            # Normalise matched tools: unique, sorted
            matched_tools_list = sorted(set(matched_tools_list))
            matched_tools_str = ','.join(matched_tools_list)
            matched_count = len(matched_tools_list)

            # Build GFF entry and update meta counters
            if pos not in ref_gene_set:  # Additional-only entry (not in reference)
                type_field = matched_tools_str if matched_tools_str else ''
                Non_Ref_Combined[matched_count] += 1

                # Update per-tool counters for additional-only entries
                if matched_tools_list:
                    for t in matched_tools_list:
                        per_tool_total[t] += 1
                        per_contig_per_tool[contig][t] += 1
                else:
                    # track unassigned additional entries (no tool name found)
                    per_tool_total['unassigned'] += 1
                    per_contig_per_tool[contig]['unassigned'] += 1

                if not getattr(options, 'clean', False):
                    entry = (contig + '\t' + type_field + '\t' + (data[2] if len(data) > 3 else '.') + '\t' + start + '\t' + stop + '\t.\t' + strand + '\t.\tID=Additional_Annotations;' + (additional_annotation_info if additional_annotation_info else '') + '\n')
                else:
                    entry = (contig + '\t' + type_field + '\t' + (data[2] if len(data) > 3 else '.') + '\t' + start + '\t' + stop + '\t.\t' + strand + '\t.\t' + (additional_annotation_info if additional_annotation_info else '') + '\n')

            else:
                # Reference entry
                if len(data) > 3 and data[3]:
                    info_field = data[3].replace('\n', '').strip()
                else:
                    info_field = '.'

                # Determine type and source fields
                type_field = data[2] if len(data) > 2 and data[2] else (
                    reference_annotation.split('/')[-1].split('.')[0] if reference_annotation else '.')
                source_field = data[5] if len(data) > 5 and data[5] else (
                    reference_annotation.split('/')[-1].split('.')[0] if reference_annotation else '')

                # If additional_annotation_info duplicates info_field content, remove duplicate fragments.
                filtered_additional = ''
                if additional_annotation_info:
                    add_parts = [p.strip() for p in str(additional_annotation_info).split(';') if p.strip()]
                    info_parts = [p.strip() for p in str(info_field).split(';') if p.strip() and p.strip() != '.']
                    filtered = []
                    for ap in add_parts:
                        dup = False
                        for ip in info_parts:
                            # treat duplication if exact match or obvious substring relationship
                            if ip and (ap == ip or ip in ap or ap in ip):
                                dup = True
                                break
                        if not dup:
                            filtered.append(ap)
                    filtered_additional = ';'.join(filtered)

                if not filtered_additional:
                    # Reference-only (no meaningful unique additional annotations)
                    Ref_Only += 1
                    if not getattr(options, 'clean', False):
                        entry = (
                                    contig + '\t' + source_field + '\t' + type_field + '\t' + start + '\t' + stop + '\t.\t' + strand + '\t.\tID=Original_Annotation;' + info_field + '\n')
                    else:
                        entry = (
                                    contig + '\t' + source_field + '\t' + type_field + '\t' + start + '\t' + stop + '\t.\t' + strand + '\t.\t' + info_field + '\n')
                else:
                    # Reference entry that had additional annotations (combined)
                    Ref_Combined[matched_count] += 1

                    # Update per-tool counters for reference-matched entries
                    if matched_tools_list:
                        for t in matched_tools_list:
                            ref_per_tool_total[t] += 1
                            ref_per_contig_per_tool[contig][t] += 1
                    else:
                        ref_per_tool_total['unassigned'] += 1
                        ref_per_contig_per_tool[contig]['unassigned'] += 1

                    if not getattr(options, 'clean', False):
                        entry = (
                                    contig + '\t' + source_field + '\t' + type_field + '\t' + start + '\t' + stop + '\t.\t' + strand + '\t.\tID=Original_Annotation;' + info_field + ';Matched_Annotations=' + filtered_additional + '\n')
                    else:
                        entry = (
                                    contig + '\t' + source_field + '\t' + type_field + '\t' + start + '\t' + stop + '\t.\t' + strand + '\t.\t' + info_field + '\n')

            write_out.write(entry)

    # Produce metadata output if requested
    if getattr(options, 'output_meta', False) == True:
        # Summaries
        total_ref_combined = sum(Ref_Combined.values())
        total_nonref = sum(Non_Ref_Combined.values())
        total_reference_genes = Ref_Only + total_ref_combined
        Ref_Combined_counter = Counter(Ref_Combined)
        Non_Ref_Combined_counter = Counter(Non_Ref_Combined)
        with open(output_file.replace('.gff','_Meta.txt'),'w') as meta_out:
            meta_out.write("GFF-Adder Metadata Report\n")
            meta_out.write("=========================\n")
            meta_out.write("Run Date: {}\n".format(date.today()))
            meta_out.write("Genome DNA: {}\n".format(getattr(options, 'genome_DNA', 'N/A')))
            meta_out.write("Reference annotation: {}\n".format(reference_annotation))
            meta_out.write("Additional annotation(s): {}\n\n".format(additional_annotation))

            meta_out.write("Summary counts\n")
            meta_out.write("--------------\n")
            meta_out.write(f"Reference-only genes (no matching additional annotation): {Ref_Only}\n")
            meta_out.write(f"Reference genes with additional matches (combined): {total_ref_combined}\n")
            meta_out.write(f"TOTAL reference genes observed: {total_reference_genes}\n")
            meta_out.write(f"Additional-only genes (not present in reference): {total_nonref}\n\n")

            meta_out.write("Distribution of matches for reference-combined entries (num matched tools -> count)\n")
            if Ref_Combined_counter:
                for k in sorted(Ref_Combined_counter):
                    meta_out.write(f"  {k:>3}    {Ref_Combined_counter[k]}\n")
            else:
                meta_out.write("  None\n")
            meta_out.write("\n")

            meta_out.write("Distribution of matches for non-reference entries (num matched tools -> count)\n")
            if Non_Ref_Combined_counter:
                for k in sorted(Non_Ref_Combined_counter):
                    meta_out.write(f"  {k:>3}    {Non_Ref_Combined_counter[k]}\n")
            else:
                meta_out.write("  None\n")
            meta_out.write("\n")

            # Per-tool totals for additional-only entries
            meta_out.write("Per-tool additional-only annotation totals\n")
            meta_out.write("-----------------------------------------\n")
            if per_tool_total:
                for t, c in per_tool_total.most_common():
                    meta_out.write(f"  {t}: {c}\n")
            else:
                meta_out.write("  None\n")
            meta_out.write("\n")

            # Per-contig breakdown for additional-only entries: only contigs with additional genes
            meta_out.write("Per-contig breakdown for additional-only annotations (only contigs with additions shown)\n")
            meta_out.write("---------------------------------------------------------------------------------------\n")
            if per_contig_per_tool:
                for contig in sorted(per_contig_per_tool):
                    counter = per_contig_per_tool[contig]
                    if sum(counter.values()) == 0:
                        continue
                    meta_out.write(f"  {contig}:\n")
                    for t, c in counter.most_common():
                        meta_out.write(f"    {t}: {c}\n")
                    meta_out.write("\n")
            else:
                meta_out.write("  None\n\n")

            # Per-tool totals and per-contig breakdown for reference genes matched by additional tools
            meta_out.write("Per-tool totals for reference genes matched by additional tools\n")
            meta_out.write("---------------------------------------------------------------\n")
            if ref_per_tool_total:
                for t, c in ref_per_tool_total.most_common():
                    meta_out.write(f"  {t}: {c}\n")
            else:
                meta_out.write("  None\n")
            meta_out.write("\n")

            meta_out.write("Per-contig breakdown for reference genes matched by additional tools\n")
            meta_out.write("--------------------------------------------------------------------\n")
            if ref_per_contig_per_tool:
                for contig in sorted(ref_per_contig_per_tool):
                    counter = ref_per_contig_per_tool[contig]
                    if sum(counter.values()) == 0:
                        continue
                    meta_out.write(f"  {contig}:\n")
                    for t, c in counter.most_common():
                        meta_out.write(f"    {t}: {c}\n")
                    meta_out.write("\n")
            else:
                meta_out.write("  None\n\n")

            meta_out.write("Notes\n")
            meta_out.write("-----\n")
            meta_out.write(" - 'Reference-only' are reference entries that had no recorded additional annotation information.\n")
            meta_out.write(" - 'Per-tool' counts are based on the tool names extracted from the additional-annotation provenance\n")
            meta_out.write("   (expected format in combined entries: 'ToolName:info'). Entries with no tool detected are shown as 'unassigned'.\n")
            meta_out.write("\nEnd of report\n")



def gff_adder(options):
    # Load fasta into dna_regions (supports multi-contig)
    try:
        try:
            fasta_in = gzip.open(options.genome_DNA, 'rt')
            dna_regions = fasta_load(fasta_in)
        except Exception:
            fasta_in = open(options.genome_DNA, 'r', encoding='unicode_escape')
            dna_regions = fasta_load(fasta_in)
    except Exception:
        # Fallback to legacy single-contig behaviour
        genome_seq = ""
        with open(options.genome_DNA, 'r') as genome_fasta:
            for line in genome_fasta:
                line = line.replace("\n", "")
                if not line.startswith('>'):
                    genome_seq += str(line)
                else:
                    genome_ID = line.split()[0].replace('>','')
        # Create dna_regions with single entry
        dna_regions = OrderedDict()
        dna_regions[genome_ID] = (genome_seq, len(genome_seq), list(), None)

    ###########################################
    # Build reference gene dict per-contig
    ref_genes_by_contig = defaultdict(OrderedDict)

    if not options.reference_tool:  # IF using Ensembl/file for comparison
        # Parse reference gff to populate ref_genes_by_contig (retain original info fields)
        # Detect gzip by magic bytes (first two bytes)
        is_gz = False
        with open(options.reference_annotation, 'rb') as _probe:
            magic = _probe.read(2)
            is_gz = (magic == b'\x1f\x8b')

        if is_gz:
            gff_in = gzip.open(options.reference_annotation, 'rt', errors='replace')
        else:
            # Open as plain text, replace undecodable bytes rather than fail
            gff_in = open(options.reference_annotation, 'r', encoding='utf-8', errors='replace')

        count = 0
        try:
            for line in gff_in:
                if line.startswith('#') or line.strip() == '':
                    continue
                parts = line.strip().split('\t')
                if len(parts) < 9:
                    continue
                contig = parts[0]
                if contig not in dna_regions:
                    # skip records for contigs not in provided fasta
                    continue
                try:
                    if 'CDS' in options.gene_ident and len(options.gene_ident) == 1:
                        if "CDS" in parts[2] and len(parts) == 9:
                            start = int(parts[3])
                            stop = int(parts[4])
                            strand = parts[6]
                            pos = f"{start},{stop}"
                            # store as [strand, source, type, info] to match downstream expectations
                            info = parts[8]
                            ref_genes_by_contig[contig][pos] = [strand, parts[1], parts[2], info]
                            count += 1
                    else:
                        gene_types = options.gene_ident.split(',')
                        if any(gene_type in parts[2] for gene_type in gene_types):
                            start = int(parts[3])
                            stop = int(parts[4])
                            strand = parts[6]
                            pos = f"{start},{stop}"
                            # store as [strand, source, type, info] to match downstream expectations
                            info = parts[8]
                            ref_genes_by_contig[contig][pos] = [strand, parts[1], parts[2], info]
                            count += 1
                except IndexError:
                    continue
        finally:
            try:
                gff_in.close()
            except Exception:
                pass

    else:
        # Reference tool provided: attempt to call it with dna_regions first (multi-contig aware), fallback to legacy signature
        reference_tool = options.reference_tool if options.reference_tool != 'StORF-Reporter' else 'StORF-Reporter'
        try:
            reference_tool_mod = import_module('Tools.' + reference_tool + '.' + reference_tool, package='my_current_pkg')
        except ModuleNotFoundError:
            try:
                reference_tool_mod = import_module('ORForise.Tools.' + reference_tool + '.' + reference_tool, package='my_current_pkg')
            except ModuleNotFoundError:
                sys.exit("Tool not available")
        reference_tool_func = getattr(reference_tool_mod, reference_tool)
        # Try multi-contig signature
        try:
            ref_result = reference_tool_func(options.reference_annotation, dna_regions)
        except TypeError:
            # Fallback to legacy signature, try passing genome seq string
            genome_seq = ''.join([dna_regions[c][0] for c in dna_regions])
            ref_result = reference_tool_func(reference_annotation=options.reference_annotation, genome_seq=genome_seq, gene_ident=options.gene_ident)
        # Expect ref_result to be dict of contig -> {pos: data}
        for contig, mapping in ref_result.items() if isinstance(ref_result, dict) else []:
            ref_genes_by_contig[contig].update(mapping)

    # Ensure each contig has an OrderedDict even if empty
    for contig in dna_regions:
        if contig not in ref_genes_by_contig:
            ref_genes_by_contig[contig] = OrderedDict()

    # Collect additional annotations per contig
    additional_annotations_by_contig = defaultdict(OrderedDict)
    tool_count = 0
    for tool in options.additional_tool.split(','):
        try:
            additional_tool_mod = import_module('Tools.' + tool + '.' + tool, package='my_current_pkg')
        except ModuleNotFoundError:
            try:
                additional_tool_mod = import_module('ORForise.Tools.' + tool + '.' + tool, package='my_current_pkg')
            except ModuleNotFoundError:
                sys.exit("Tool not available")
        additional_tool_func = getattr(additional_tool_mod, tool)

        anno_file = options.additional_annotation.split(',')[tool_count]
        tool_count += 1
        # Try calling tool in multi-contig mode first
        try:
            tool_orfs = additional_tool_func(anno_file, dna_regions)
        except TypeError:
            # Fallback to legacy signature expecting genome_seq
            genome_seq = ''.join([dna_regions[c][0] for c in dna_regions])
            tool_orfs = additional_tool_func(anno_file, genome_seq, options.gene_ident)

        # tool_orfs may be either {contig: {pos: data}} or a flat {pos: data}
        if isinstance(tool_orfs, dict):
            # If top-level keys look like contig names (present in dna_regions) then treat as multi-contig
            top_keys = list(tool_orfs.keys())
            if top_keys and top_keys[0] in dna_regions:
                for contig, mapping in tool_orfs.items():
                    # Merge mapping into additional_annotations_by_contig and record tool provenance
                    for pos_k, pos_v in mapping.items():
                        # store tuple (value, tool)
                        additional_annotations_by_contig[contig][pos_k] = (pos_v, tool)
            else:
                # Treat as flat mapping — assume single contig if only one contig present
                if len(dna_regions) == 1:
                    only_contig = next(iter(dna_regions))
                    for pos_k, pos_v in tool_orfs.items():
                        additional_annotations_by_contig[only_contig][pos_k] = (pos_v, tool)
                else:
                    # If multiple contigs but mapping has contig-prefixed keys like 'contig,start,stop', split them
                    for k, v in tool_orfs.items():
                        parts = k.split(',')
                        if len(parts) == 3 and parts[0] in dna_regions:
                            contig = parts[0]
                            pos = parts[1] + ',' + parts[2]
                            additional_annotations_by_contig[contig][pos] = (v, tool)
                        else:
                            # Unknown format: assign nowhere (skip)
                            continue
        tool_orfs = None

    # Combine per-contig: keep reference entries and append additional annotations as supplemental
    combined_ORFs_by_contig = OrderedDict()
    for contig in dna_regions:
        combined = OrderedDict()
        # Add reference entries first; normalise to [strand, 'ref', type, ref_info, additional_list]
        for pos, val in ref_genes_by_contig.get(contig, {}).items():
            strand = val[0] if len(val) > 0 else '.'
            src = 'ref'
            source_field = val[1] if len(val) > 1 else 'ref'
            ftype = val[2] if len(val) > 2 else '.'
            ref_info = val[3] if len(val) > 3 else '.'
            combined[pos] = [strand, src, ftype, ref_info, [], source_field]

        # Now incorporate additional annotations without overwriting reference entries
        for pos, wrapped in additional_annotations_by_contig.get(contig, {}).items():
            # wrapped is (value, tool)
            if isinstance(wrapped, tuple) and len(wrapped) == 2:
                val, toolname = wrapped
            else:
                val = wrapped
                toolname = ''

            # Extract strand/type/info from value heuristically
            strand_a = val[0] if isinstance(val, (list, tuple)) and len(val) > 0 else '.'
            ftype_a = val[3] if isinstance(val, (list, tuple)) and len(val) > 2 else '.'
            info_a = ''
            if isinstance(val, (list, tuple)) and len(val) > 3:
                info_a = val[4]
            elif isinstance(val, str):
                info_a = val

            # If matching pos exists in reference, append additional info to its additional list
            if pos in combined:
                addstr = (toolname + ':' + info_a) if toolname else info_a
                combined[pos][4].append(addstr)
            else:
                # Create a new entry for additional-only annotation: [strand, 'add', type, '.', [tool:info]]
                addstr = (toolname + ':' + info_a) if toolname else info_a
                #combined[pos] = [strand_a, 'add', ftype_a if ftype_a else '.', '.', [addstr]]
                combined[pos] = [strand_a, 'add', ftype_a if ftype_a else '.', '.', [addstr], toolname]
        # Sort ORFs for this contig
        combined = sortORFs(combined)
        combined_ORFs_by_contig[contig] = combined

    # Call writer
    gff_writer(options, combined_ORFs_by_contig, options.output_file, options.reference_annotation, options.additional_annotation)


def main():
    parser = argparse.ArgumentParser(description='ORForise ' + ORForise_Version + ': GFF-Adder Run Parameters.')
    parser._action_groups.pop()

    required = parser.add_argument_group('Required Arguments')
    required.add_argument('-dna', dest='genome_DNA', required=True, help='Genome DNA file (.fa) which both annotations '
                                                                    'are based on')
    required.add_argument('-ref', dest='reference_annotation', required=True,
                        help='Which reference annotation file to use as reference?')
    required.add_argument('-at', dest='additional_tool', required=True,
                        help='Which format to use for additional annotation? - Can provide multiple annotations (Tool1,Tool2)')
    required.add_argument('-add', dest='additional_annotation', required=True,
                        help='Which annotation file to add to reference annotation? - Can provide multiple annotations (1.GFF,2.GFF)')
    required.add_argument('-o', dest='output_file', required=True,
                        help='Output filename')

    optional = parser.add_argument_group('Optional Arguments')
    optional.add_argument('-rt', dest='reference_tool', required=False,
                        help='Which tool format to use as reference? - If not provided, will default to the '
                             'standard GFF format and will only look for "CDS" features')
    optional.add_argument('--gene_ident', action='store', dest='gene_ident', default='CDS',
                        help='Identifier used for identifying genomic features in reference annotation "CDS,rRNA,tRNA"')
    optional.add_argument('-mc', dest='mark_consensus', action='store_true', required=False,
                        help='Default - False: Mark reference annotations which where present in the additional tool annotation')
    optional.add_argument('-c', dest='clean', action='store_true', required=False,
                        help='Default - False: Do not mark 9th column with "Original/Matched/Additional tag"')
    optional.add_argument('--meta', dest='output_meta', action='store_true', required=False,
                        help='Default - False: Output metadata file')
    optional.add_argument('--olap', dest='overlap', default=50, type=int, required=False,
                        help='Maximum overlap between reference and additional genic regions (CDS,rRNA etc) - Default: 50 nt')

    misc = parser.add_argument_group('Misc')
    misc.add_argument('-v', dest='verbose', default='False', type=eval, choices=[True, False],
                      help='Default - False: Print out runtime status')

    options = parser.parse_args()

    gff_adder(options)



if __name__ == "__main__":
    try:
        try:
            main()
        except Exception:
            logging.exception('Unhandled exception in main')
    finally:
        print(CLOSING)


from importlib import import_module
import argparse
from collections import OrderedDict
from datetime import date
import sys, gzip
import os
import logging

# Ensure logging prints to stdout by default so info/debug messages are visible when running the script
if not logging.getLogger().handlers:
    logging.basicConfig(stream=sys.stdout, level=logging.INFO, format='%(levelname)s: %(message)s')

try:
    from .utils import *
except (ImportError, ModuleNotFoundError):
    from utils import *

################################


def gff_writer(genome_ID, genome_DNA, reference_annotation, reference_tool, ref_gene_set, additional_annotation, additional_tool, genes_To_Keep_by_contig, output_file, gene_ident=None):
    # genes_To_Keep_by_contig: {contig: {pos: data}}
    # Expand user (~) and ensure output directory exists
    output_file = os.path.expanduser(output_file)
    out_dir = os.path.dirname(output_file)
    if out_dir:
        os.makedirs(out_dir, exist_ok=True)

    try:
        with open(output_file, 'w', encoding='utf-8') as write_out:
            write_out.write("##gff-version\t3\n#\tAnnotation-Intersector\n#\tRun Date:" + str(date.today()) + '\n')
            if genome_DNA:
                write_out.write("##Genome DNA File:" + genome_DNA + '\n')
            write_out.write("##Original File: " + (reference_annotation or '') + "\n##Intersecting File: " + (additional_annotation or '') + '\n')

            entries_written = 0

            # If genes_To_Keep_by_contig is falsy or empty, attempt to write reference features directly as fallback
            if not genes_To_Keep_by_contig or all(len(v) == 0 for v in genes_To_Keep_by_contig.values()):
                write_out.write(f"# No kept genes to write (0 entries).\n")
                write_out.write(f"# Falling back to writing reference features with coverage=0.\n")

                # Parse reference annotation and write features matching gene_ident
                try:
                    if reference_annotation and reference_annotation.endswith('.gz'):
                        rf = gzip.open(reference_annotation, 'rt')
                    else:
                        rf = open(reference_annotation, 'r', encoding='unicode_escape')
                    with rf:
                        for line in rf:
                            line = line.rstrip('\n')
                            if not line or line.startswith('#'):
                                continue
                            parts = line.split('\t')
                            if len(parts) < 9:
                                continue
                            seqid = parts[0]
                            ftype = parts[2]
                            try:
                                gene_types = gene_ident.split(',') if gene_ident else ['CDS']
                            except Exception:
                                gene_types = ['CDS']
                            if ftype not in gene_types and not ('CDS' in gene_types and ftype == 'CDS'):
                                continue
                            try:
                                start = parts[3]
                                stop = parts[4]
                                strand = parts[6]
                                info = parts[8]
                                source = parts[1] if len(parts) > 1 else ''
                            except Exception:
                                continue
                            # write entry with coverage 0 and empty additional annotation; use source from input GFF
                            entry = f"{seqid}\t{source}\t{ftype}\t{start}\t{stop}\t.\t{strand}\t.\tID=Original_Annotation={info}\n"
                            write_out.write(entry)
                            entries_written += 1
                except Exception as e:
                    logging.warning('Fallback parse of reference annotation failed: %s', e)

                write_out.flush()
                logging.info('Wrote %d fallback reference entries to %s', entries_written, output_file)
                return

            # Iterate contigs and write kept entries. Kept entry layout expected:
            # [strand, coverage, additional_type, ref_type, additional_info, ref_info, ref_source]
            for contig, genes in genes_To_Keep_by_contig.items():
                fallback_source = os.path.splitext(os.path.basename(reference_annotation))[0].split('_')[0] if reference_annotation else 'reference'
                for pos, data in genes.items():
                    try:
                        pos_ = pos.split(',')
                        start = pos_[0]
                        stop = pos_[-1]
                        strand = data[0] if len(data) > 0 else '.'
                        add_ann = str(data[4]) if len(data) > 4 else ''
                        orig_ann = str(data[5]) if len(data) > 5 else ''
                        source_field = data[6] if len(data) > 6 and data[6] else fallback_source
                        feat_type = data[3] if len(data) > 3 and data[3] else (data[2] if len(data) > 2 else 'CDS')
                        entry = (
                            contig + '\t' + source_field + '\t' + feat_type + '\t' + start + '\t' + stop + '\t.\t' + strand + '\t.\tID=Original_Annotation=' + orig_ann + ';Additional_Annotation=' + add_ann + ';Coverage=' + str(
                                data[1]) + '\n')
                        write_out.write(entry)
                        entries_written += 1
                    except Exception as e:
                        logging.warning('Skipping bad GFF entry for contig %s pos %s: %s', contig, pos, e)
                        continue

            write_out.flush()
            logging.info('Wrote %d GFF entries to %s', entries_written, output_file)
    except OSError as e:
        logging.error("Cannot write to output file %s: %s", output_file, e)
        sys.exit(1)


def _get_opt(options, *names):

    for n in names:
        if hasattr(options, n):
            return getattr(options, n)
    return None


def _parse_pos(pos_str):
    try:
        s, e = pos_str.split(',')
        return int(s), int(e)
    except Exception:
        return None, None


def _write_discordance_report(report_path, entries):
    # Summarise discordance entries instead of writing each row (GFFs keep the full detail)
    report_path = os.path.expanduser(report_path)
    out_dir = os.path.dirname(report_path)
    if out_dir:
        os.makedirs(out_dir, exist_ok=True)
    try:
        from collections import Counter
        total = len(entries) if entries is not None else 0
        status_counts = Counter()
        contig_counts = Counter()
        cov_values = []
        for e in (entries or []):
            st = str(e.get('status', 'unknown'))
            status_counts[st] += 1
            contig = str(e.get('contig', ''))
            if contig:
                contig_counts[contig] += 1
            # attempt to extract numeric coverage if present
            try:
                cov = float(e.get('coverage', 0) if e.get('coverage', '') != '' else 0)
                cov_values.append(cov)
            except Exception:
                try:
                    # sometimes coverage might be a string like '12.34' or '0.00'
                    cov_values.append(float(str(e.get('coverage', '0')).strip()))
                except Exception:
                    pass

        avg_cov = (sum(cov_values) / len(cov_values)) if cov_values else 0.0
        nonzero_covs = [v for v in cov_values if v != 0]
        nonzero_avg_cov = (sum(nonzero_covs) / len(nonzero_covs)) if nonzero_covs else 0.0

        with open(report_path, 'w', encoding='utf-8') as fh:
            fh.write('metric\tvalue\n')
            fh.write(f'total_entries\t{total}\n')
            fh.write(f'unique_contigs_with_discordance\t{len(contig_counts)}\n')
            fh.write(f'average_coverage_reported\t{avg_cov:.2f}\n')
            fh.write(f'non-zero_average_coverage_reported\t{nonzero_avg_cov:.2f}\n')
            # statuses
            for status, cnt in status_counts.most_common():
                fh.write(f'status::{status}\t{cnt}\n')
            # per-contig counts (only top 50 to avoid huge files)
            fh.write('# per-contig discordance counts (top 50)\n')
            for contig, cnt in contig_counts.most_common(50):
                fh.write(f'contig::{contig}\t{cnt}\n')
        logging.info('Wrote discordance summary report: %s', report_path)
    except OSError as e:
        logging.error('Cannot write discordance report %s: %s', report_path, e)
        sys.exit(1)


def _write_discordance_gff(report_path, entries, reference_annotation_basename):
    """Write a list of discordance entries (dicts) to a GFF file.
    The GFF source column is taken from carried 'ref_source' or 'add_source' when available.
    """
    report_path = os.path.expanduser(report_path)
    out_dir = os.path.dirname(report_path)
    if out_dir:
        os.makedirs(out_dir, exist_ok=True)
    try:
        with open(report_path, 'w', encoding='utf-8') as fh:
            fh.write('##gff-version\t3\n')
            fh.write('#\tAnnotation-Intersector discordance report\n')
            fh.write('#\tRun Date:' + str(date.today()) + '\n')
            fh.write('##Original File: ' + (reference_annotation_basename or '') + '\n')
            entries_written = 0
            for e in (entries or []):
                try:
                    contig = str(e.get('contig', '.'))
                    ref_pos = e.get('ref_pos', '')
                    add_pos = e.get('add_pos', '')
                    if ref_pos:
                        start, stop = ref_pos.split(',')
                        ftype = e.get('ref_type', '') or 'CDS'
                        source = e.get('ref_source') or (reference_annotation_basename.split('_')[0] if reference_annotation_basename else 'reference')
                        info_attr = e.get('ref_info', '')
                    else:
                        start, stop = add_pos.split(',') if add_pos else ('0', '0')
                        ftype = e.get('add_type', '') or 'CDS'
                        source = e.get('add_source') or 'additional'
                        info_attr = e.get('add_info', '')
                    attrs = []
                    attrs.append('Status=' + str(e.get('status', '')))
                    attrs.append('Coverage=' + str(e.get('coverage', '')))
                    if e.get('ref_info', ''):
                        attrs.append('Ref_info=' + str(e.get('ref_info', '')).replace(';', '%3B'))
                    if e.get('add_info', ''):
                        attrs.append('Add_info=' + str(e.get('add_info', '')).replace(';', '%3B'))
                    attr_str = ';'.join(attrs)
                    line = f"{contig}\t{source}\t{ftype}\t{start}\t{stop}\t.\t.\t.\t{attr_str}\n"
                    fh.write(line)
                    entries_written += 1
                except Exception:
                    continue
        logging.info('Wrote %d discordance GFF entries to %s', entries_written, report_path)
    except OSError as e:
        logging.error('Cannot write discordance GFF %s: %s', report_path, e)
        sys.exit(1)


def compute_discordance(ref_map_by_contig, add_map_by_contig, options):
    """Compare reference and additional maps per contig and return three lists:
    - only_in_ref: reference entries with no overlapping additional ORF
    - only_in_additional: additional ORFs that don't overlap any reference entry
    - mismatches: reference entries with overlapping additional ORFs that don't meet match criteria

    Expected layouts:
    - ref entry: [strand, 'ref', type, info, source]
    - add entry: [strand, ..., type (index 3), info (last element), (optional) source]
    """
    only_in_ref = []
    only_in_additional = []
    mismatches = []

    all_contigs = list(OrderedDict.fromkeys(list(ref_map_by_contig.keys()) + list(add_map_by_contig.keys())))
    matched_adds = set()
    cov_thresh = getattr(options, 'coverage', 100.0)

    for contig in all_contigs:
        ref_map = ref_map_by_contig.get(contig, {}) or {}
        add_map = add_map_by_contig.get(contig, {}) or {}

        for rpos, rdata in ref_map.items():
            rstart, rstop = _parse_pos(rpos)
            if rstart is None:
                continue
            rlen = (rstop - rstart + 1) if (rstop and rstart) else 0
            best_cov = 0.0
            best_add = None
            best_add_data = None
            matched = False

            r_strand = rdata[0] if len(rdata) > 0 else ''
            r_type = rdata[2] if len(rdata) > 2 else ''
            r_info = rdata[3] if len(rdata) > 3 else ''
            r_source = rdata[4] if len(rdata) > 4 else ''

            for apos, adata in add_map.items():
                astart, astop = _parse_pos(apos)
                if astart is None:
                    continue
                ov = max(0, min(rstop, astop) - max(rstart, astart) + 1)
                if ov <= 0:
                    continue
                cov = 100.0 * float(ov) / float(rlen) if rlen > 0 else 0.0
                if cov > best_cov:
                    best_cov = cov
                    best_add = apos
                    best_add_data = adata

                a_strand = adata[0] if len(adata) > 0 else ''
                a_type = adata[3] if len(adata) > 3 else ''
                a_info = adata[-1] if len(adata) > 0 else ''
                a_source = adata[4] if len(adata) > 4 else ''

                try:
                    frame_ok = ((abs(astop - rstop) % 3) == 0)
                except Exception:
                    frame_ok = True

                if a_type == r_type and cov >= cov_thresh and (a_strand == r_strand) and frame_ok:
                    matched = True
                    matched_adds.add((contig, apos))
                    break

            if matched:
                continue

            if best_add is None:
                only_in_ref.append({
                    'contig': contig,
                    'ref_pos': rpos,
                    'add_pos': '',
                    'ref_type': r_type,
                    'add_type': '',
                    'status': 'only_in_ref',
                    'coverage': '0.00',
                    'ref_info': r_info,
                    'ref_source': r_source,
                    'add_info': ''
                })
            else:
                a_type = best_add_data[3] if len(best_add_data) > 3 else ''
                a_info = best_add_data[-1] if len(best_add_data) > 0 else ''
                a_strand = best_add_data[0] if len(best_add_data) > 0 else ''
                a_source = best_add_data[4] if len(best_add_data) > 4 else ''

                type_match = (a_type == r_type)
                strand_match = (a_strand == r_strand)
                cov_ok = (best_cov >= cov_thresh)
                try:
                    astart, astop = _parse_pos(best_add)
                    frame_ok = ((abs(astop - rstop) % 3) == 0) if (astop is not None) else True
                except Exception:
                    frame_ok = True

                if not cov_ok:
                    status = 'found_in_additional_but_below_coverage'
                elif not type_match and not strand_match:
                    status = 'found_in_additional_different_type_and_strand'
                elif not type_match:
                    status = 'found_in_additional_different_type'
                elif not strand_match:
                    status = 'found_in_additional_different_strand'
                elif not frame_ok:
                    status = 'found_in_additional_different_frame'
                else:
                    status = 'partial_overlap'

                mismatches.append({
                    'contig': contig,
                    'ref_pos': rpos,
                    'add_pos': best_add or '',
                    'ref_type': r_type,
                    'add_type': a_type,
                    'status': status,
                    'coverage': f"{best_cov:.2f}",
                    'ref_info': r_info,
                    'ref_source': r_source,
                    'add_info': a_info,
                    'add_source': a_source,
                })

                if best_add:
                    matched_adds.add((contig, best_add))

        for apos, adata in add_map.items():
            if (contig, apos) in matched_adds:
                continue
            astart, astop = _parse_pos(apos)
            if astart is None:
                continue
            overlapped = False
            for rpos in ref_map.keys():
                rstart, rstop = _parse_pos(rpos)
                if rstart is None:
                    continue
                if max(rstart, astart) <= min(rstop, astop):
                    overlapped = True
                    break
            if not overlapped:
                add_source = adata[4] if len(adata) > 4 else ''
                only_in_additional.append({
                    'contig': contig,
                    'ref_pos': '',
                    'add_pos': apos,
                    'ref_type': '',
                    'add_type': adata[3] if len(adata) > 3 else '',
                    'status': 'only_in_additional',
                    'coverage': '0.00',
                    'ref_info': '',
                    'add_info': adata[-1] if len(adata) > 0 else '',
                    'add_source': add_source,
                })

    # Return discordance lists and the set of matched additional ORFs (for overlap counts)
    return only_in_ref, only_in_additional, mismatches, matched_adds


def comparator(options):
    # Multi-contig aware comparator
    genome_seq = ''
    genome_ID = None
    dna_regions = {}

    # Support both 'genome_DNA' and 'genome_dna' option names (compat with Annotation_Compare)
    genome_path = _get_opt(options, 'genome_DNA', 'genome_dna')

    # Load genome fasta if provided
    if genome_path:
        if not os.path.exists(genome_path):
            logging.error('Genome DNA file does not exist: %s', genome_path)
            sys.exit(1)
        try:
            fasta_in = gzip.open(genome_path, 'rt')
            dna_regions = fasta_load(fasta_in)
        except Exception:
            fasta_in = open(genome_path, 'r', encoding='unicode_escape')
            dna_regions = fasta_load(fasta_in)
        # genome_ID fallback
        try:
            if isinstance(dna_regions, dict) and len(dna_regions) > 0:
                genome_ID = next(iter(dna_regions.keys()))
                genome_seq = dna_regions[genome_ID]
            else:
                genome_ID = os.path.splitext(os.path.basename(genome_path))[0]
        except Exception:
            genome_seq = ''
            genome_ID = os.path.splitext(os.path.basename(genome_path))[0]
    else:
        # derive genome_ID from reference annotation filename
        genome_seq = ''
        genome_ID = os.path.splitext(os.path.basename(options.reference_annotation))[0]


    # Load reference annotation. If a tool-specific parser is requested, use it to ensure contig keys exist in dna_regions
    if getattr(options, 'reference_tool', None):
        try:
            reference_tool_mod = import_module('Tools.' + options.reference_tool + '.' + options.reference_tool,
                                               package='my_current_pkg')
        except ModuleNotFoundError:
            try:
                reference_tool_mod = import_module(
                    'ORForise.Tools.' + options.reference_tool + '.' + options.reference_tool,
                    package='my_current_pkg')
            except ModuleNotFoundError:
                logging.error('Reference tool module not available: %s', options.reference_tool)
                sys.exit(1)
        reference_tool_fn = getattr(reference_tool_mod, options.reference_tool)
        try:
            # Call the tool parser; many tools return a contig->dict mapping. Ensure dna_regions contains those contig keys.
            ref_out = reference_tool_fn(options.reference_annotation, dna_regions)
            if isinstance(ref_out, dict):
                for contig_key in ref_out.keys():
                    if contig_key not in dna_regions:
                        dna_regions[contig_key] = ['']
        except Exception as e:
            logging.error('Failed to load reference annotation with tool %s: %s', options.reference_tool, e)
            sys.exit(1)
    else:
        try:
            gff_in = gzip.open(options.reference_annotation, 'rt')
            dna_regions = gff_load(options, gff_in, dna_regions)
        except Exception:
            gff_in = open(options.reference_annotation, 'r', encoding='unicode_escape')
            dna_regions = gff_load(options, gff_in, dna_regions)

    # Build ref_genes_by_contig: mapping contig -> OrderedDict(pos -> [strand, 'ref', type, info])
    ref_genes_by_contig = OrderedDict()

    if not getattr(options, 'reference_tool', None):
        # Parse GFF and group by seqid (first column)
        with open(options.reference_annotation, 'r', encoding='unicode_escape') as genome_gff:
            for line in genome_gff:
                line = line.rstrip('\n')
                if not line or line.startswith('#'):
                    continue
                parts = line.split('\t')
                if len(parts) < 9:
                    continue
                seqid = parts[0]
                ftype = parts[2]
                try:
                    # Determine if this feature type is requested
                    gene_types = options.gene_ident.split(',') if options.gene_ident else ['CDS']
                except Exception:
                    gene_types = ['CDS']
                if ftype not in gene_types and not ('CDS' in gene_types and ftype == 'CDS'):
                    # If user specified CDS and this is CDS, include; else skip
                    if ftype not in gene_types:
                        continue
                try:
                    start = int(parts[3])
                    stop = int(parts[4])
                    strand = parts[6]
                    pos = f"{start},{stop}"
                    info = parts[8]
                    source = parts[1] if len(parts) > 1 else ''
                except (IndexError, ValueError):
                    continue
                if seqid not in ref_genes_by_contig:
                    ref_genes_by_contig[seqid] = OrderedDict()
                # Store source from column 1 as well. Layout becomes:
                # [strand, 'ref', type, info, source]
                ref_genes_by_contig[seqid].update({pos: [strand, 'ref', ftype, info, source]})
    else:
        # Use a tool parser to produce ref_genes; expect tool to return mapping contig->dict
        try:
            reference_tool_mod = import_module('Tools.' + options.reference_tool + '.' + options.reference_tool,
                                               package='my_current_pkg')
        except ModuleNotFoundError:
            try:
                reference_tool_mod = import_module('ORForise.Tools.' + options.reference_tool + '.' + options.reference_tool,
                                                   package='my_current_pkg')
            except ModuleNotFoundError:
                sys.exit("Tool not available")
        reference_tool_fn = getattr(reference_tool_mod, options.reference_tool)
        ref_out = reference_tool_fn(options.reference_annotation, dna_regions)
        # If the tool returns a mapping contig->dict, use that; otherwise assume single-contig and wrap
        if isinstance(ref_out, dict) and any(isinstance(v, dict) for v in ref_out.values()):
            ref_genes_by_contig = ref_out
        else:
            # single-contig output: place under genome_ID or first contig in dna_regions
            contig_key = genome_ID if genome_ID in dna_regions else (next(iter(dna_regions.keys())) if dna_regions else genome_ID)
            ref_genes_by_contig[contig_key] = ref_out

    # Get additional ORFs using tool parser; expect mapping contig->dict
    try:
        additional_tool_mod = import_module('Tools.' + options.additional_tool + '.' + options.additional_tool,
                                            package='my_current_pkg')
    except ModuleNotFoundError:
        try:
            additional_tool_mod = import_module('ORForise.Tools.' + options.additional_tool + '.' + options.additional_tool,
                                                package='my_current_pkg')
        except ModuleNotFoundError:
            sys.exit("Tool not available")
    additional_tool_fn = getattr(additional_tool_mod, options.additional_tool)
    additional_orfs = additional_tool_fn(options.additional_annotation, dna_regions)

    # Normalise additional_orfs: if single-contig dict, wrap under appropriate contig key
    if isinstance(additional_orfs, dict) and any(isinstance(v, dict) for v in additional_orfs.values()):
        additional_by_contig = additional_orfs
    else:
        contig_key = genome_ID if genome_ID in dna_regions else (next(iter(dna_regions.keys())) if dna_regions else genome_ID)
        additional_by_contig = {contig_key: additional_orfs}

    genes_To_Keep_by_contig = OrderedDict()

    # Iterate per contig and perform intersection logic
    for contig, orfs in additional_by_contig.items():
        ref_genes = ref_genes_by_contig.get(contig, OrderedDict())
        kept = OrderedDict()
        if options.coverage == 100.00:
            for orf, data in orfs.items():
                try:
                    o_Start = int(orf.split(',')[0])
                    o_Stop = int(orf.split(',')[1])
                except Exception:
                    continue
                o_Strand = data[0]
                additional_type = data[3]
                additional_info = data[-1]

                # Lookup exact-match reference entry safely
                ref_entry = ref_genes.get(f"{o_Start},{o_Stop}")
                if not ref_entry:
                    continue
                # ref_entry layout now: [strand, 'ref', type, info, source]
                ref_type = ref_entry[2] if len(ref_entry) > 2 else ''
                ref_info = ref_entry[3] if len(ref_entry) > 3 else ''
                ref_source = ref_entry[4] if len(ref_entry) > 4 else ''

                if additional_type == ref_type and o_Strand == ref_entry[0]:
                    # kept layout: [strand, coverage, additional_type, ref_type, additional_info, ref_info, ref_source]
                    kept.update({f"{o_Start},{o_Stop}": [o_Strand, options.coverage, additional_type, ref_type, additional_info, ref_info, ref_source]})
        else:
            cov_thresh = options.coverage
            for orf, data in orfs.items():
                try:
                    o_Start = int(orf.split(',')[0])
                    o_Stop = int(orf.split(',')[1])
                except Exception:
                    continue
                o_Strand = data[0]
                additional_type = data[3]
                additional_info = data[-1]

                for gene, r_data in ref_genes.items():
                    try:
                        g_Start = int(gene.split(',')[0])
                        g_Stop = int(gene.split(',')[1])
                    except Exception:
                        continue

                    # skip genes that start after this ORF (ref genes assumed sorted by start)
                    if g_Start > o_Stop:
                        break
                    # skip genes that end before this ORF
                    if g_Stop < o_Start:
                        continue

                    # compute overlap length without creating large sets
                    overlap = max(0, min(o_Stop, g_Stop) - max(o_Start, g_Start) + 1)
                    gene_len = (g_Stop - g_Start + 1)
                    if gene_len <= 0:
                        continue
                    cov = 100.0 * overlap / gene_len

                    g_Strand = r_data[0]
                    # r_data layout now: [strand, 'ref', type, info, source]
                    ref_type = r_data[2] if len(r_data) > 2 else ''
                    ref_info = r_data[3] if len(r_data) > 3 else ''
                    ref_source = r_data[4] if len(r_data) > 4 else ''

                    if abs(o_Stop - g_Stop) % 3 == 0 and o_Strand == g_Strand and cov >= cov_thresh:
                        if additional_type == ref_type:
                            # keep ref_source with the kept entry
                            kept[f"{g_Start},{g_Stop}"] = [g_Strand, int(cov), additional_type, ref_type, additional_info, ref_info, ref_source]
        genes_To_Keep_by_contig[contig] = sortORFs(kept)

    # Log counts for debugging why GFF might be empty
    # Compute summary metrics (safe/simple)
    total_ref = sum(len(v) for v in ref_genes_by_contig.values()) if ref_genes_by_contig else 0
    total_add = sum(len(v) for v in additional_by_contig.values()) if additional_by_contig else 0
    total_kept = sum(len(v) for v in genes_To_Keep_by_contig.values()) if genes_To_Keep_by_contig else 0
    # Print totals in requested order: reference, additional, then overlap/kept and a percentage
    logging.info('Totals -- reference_genes=%d, additional_genes=%d, overlapping/kept=%d', total_ref, total_add, total_kept)
    if total_ref:
        logging.info('Overlap relative to reference: %.2f%%', (100.0 * total_kept / total_ref))

    # If requested, compute discordance lists and write three GFF outputs
    if getattr(options, 'report_discordance', False):
        # Compute discordance lists and matched additional ORFs
        only_in_ref, only_in_additional, mismatches, matched_adds = compute_discordance(ref_genes_by_contig, additional_by_contig, options)
        base = os.path.splitext(os.path.basename(options.output_file))[0] if getattr(options, 'output_file', None) else 'discordance'
        outdir = os.path.dirname(options.output_file) if getattr(options, 'output_file', None) else '.'
        ref_base = os.path.splitext(os.path.basename(options.reference_annotation))[0]

        # Compute and log clear summary metrics
        total_ref = sum(len(v) for v in ref_genes_by_contig.values()) if ref_genes_by_contig else 0
        total_add = sum(len(v) for v in additional_by_contig.values()) if additional_by_contig else 0
        overlapping_additional = len(matched_adds) if matched_adds is not None else 0
        overlapping_reference = max(0, total_ref - (len(only_in_ref) if only_in_ref is not None else 0))
        logging.info('Summary: reference_genes=%d, additional_geness=%d, additional_genes_overlapping_any_reference=%d, reference_genes_overlapped=%d', total_ref, total_add, overlapping_additional, overlapping_reference)

         # Keep the three detailed GFF outputs (backward compatible)
        gff_ref = os.path.join(outdir, f"{base}.only_in_reference.gff")
        gff_add = os.path.join(outdir, f"{base}.only_in_additional.gff")
        gff_mis = os.path.join(outdir, f"{base}.mismatches.gff")
        try:
            _write_discordance_gff(gff_ref, only_in_ref, ref_base)
            logging.info('Wrote discordance GFF: %s', gff_ref)
        except Exception:
            logging.exception('Failed to write discordance GFF: %s', gff_ref)
        try:
            _write_discordance_gff(gff_add, only_in_additional, ref_base)
            logging.info('Wrote discordance GFF: %s', gff_add)
        except Exception:
            logging.exception('Failed to write discordance GFF: %s', gff_add)
        try:
            _write_discordance_gff(gff_mis, mismatches, ref_base)
            logging.info('Wrote discordance GFF: %s', gff_mis)
        except Exception:
            logging.exception('Failed to write discordance GFF: %s', gff_mis)

        # Write a single concise summary TSV aggregating all discordance entries
        combined = []
        combined.extend(only_in_ref or [])
        combined.extend(mismatches or [])
        combined.extend(only_in_additional or [])
        combined_tsv = os.path.join(outdir, f"{base}.discordance_summary.tsv")
        try:
            _write_discordance_report(combined_tsv, combined)
            logging.info('Wrote discordance summary TSV: %s', combined_tsv)
        except Exception:
            logging.exception('Failed to write discordance summary TSV: %s', combined_tsv)

    # Ensure we always write a GFF (header + possibly diagnostic) so the core file is not empty
    genome_DNA_path = genome_path if genome_path else None



    # Write the kept genes GFF (this was missing and is why gff_writer wasn't called)
    try:
        logging.info('About to call gff_writer: total_kept=%d', total_kept)
        try:
            contig_summary = {c: len(v) for c, v in genes_To_Keep_by_contig.items()}
        except Exception:
            contig_summary = {}
        logging.info('Kept genes by contig (sample): %s', dict(list(contig_summary.items())[:10]))
        logging.info('Writing combined GFF to %s', options.output_file)
        # single correct invocation of gff_writer
        gff_writer(genome_ID, genome_DNA_path, options.reference_annotation, getattr(options, 'reference_tool', None), None, options.additional_annotation, options.additional_tool, genes_To_Keep_by_contig, options.output_file, getattr(options, 'gene_ident', None))
        logging.info('gff_writer finished (check output file)')
    except Exception as e:
        logging.exception('Failed to write combined GFF: %s', e)

    # End of comparator


def main():
    parser = argparse.ArgumentParser(description='ORForise ' + ORForise_Version + ': Annotation-Intersector Run Parameters')

    required = parser.add_argument_group('Required Arguments')
    required.add_argument('-ref', dest='reference_annotation', required=True,
                          help='Reference annotation GFF file')
    required.add_argument('-at', dest='additional_tool', required=True,
                          help='Tool name/format for additional annotation (module under Tools/)')
    required.add_argument('-add', dest='additional_annotation', required=True,
                          help='Additional annotation file to compare')
    required.add_argument('-o', dest='output_file', required=True,
                          help='Output GFF filename for kept genes')

    optional = parser.add_argument_group('Optional Arguments')
    optional.add_argument('-dna', dest='genome_DNA', required=False,
                          help='Genome DNA file (.fa) which both annotations are based on')
    optional.add_argument('-rt', dest='reference_tool', required=False,
                          help='Reference tool parser name (if not provided, GFF is expected)')
    optional.add_argument('-gi', dest='gene_ident', default='CDS', required=False,
                          help='Comma-separated feature types to consider from reference (default: CDS)')
    optional.add_argument('-cov', '--coverage', dest='coverage', default=100.0, type=float, required=False,
                          help='Percentage coverage threshold for intersection (default 100)')
    optional.add_argument('--report-discordance', dest='report_discordance', action='store_true', required=False,
                          help='If set, produce discordance reports (three GFFs)')
    optional.add_argument('--report-discordance-file', dest='report_discordance_file', required=False,
                          help='Optional base path for discordance reports')

    options = parser.parse_args()
    comparator(options)


if __name__ == '__main__':
    try:
        try:
            main()
        except Exception:
            logging.exception('Unhandled exception in main')
    finally:
        print(CLOSING)













#!/usr/bin/env python3
import collections

# Constants
SHORT_ORF_LENGTH = 300
MIN_COVERAGE = 75
ORForise_Version = 'v1.6.4'
CLOSING=("\n####\nThank you for using ORForise\nPlease report any issues to: https://github.com/NickJD/ORForise/issues\n"
        "Please Cite: https://doi.org/10.1093/bioinformatics/btab827\n"
        "#####")


def revCompIterative(watson):  # Gets Reverse Complement
    return watson.upper()[::-1].translate(str.maketrans("ATCGRYKMVBHD","TAGCYRMKBVDH"))


def sortORFs(tool_ORFs):  # Will only sort by given start position
    tool_ORFs_Sorted = sorted(tool_ORFs.items(), key=lambda v: int(v[0].split(",")[0]))
    tool_ORFs_Sorted = collections.OrderedDict(tool_ORFs_Sorted)
    return tool_ORFs_Sorted


def sortGenes(Genes):  # Will sort by given start position and then rearrange for given stop
    Genes_Sorted_list = sorted(Genes.values(), key=lambda v: int(v[0]))
    Genes_Sorted = []
    for idx,gene in enumerate(Genes_Sorted_list):
        Genes_Sorted.append([idx,gene])
    Genes_Sorted = collections.OrderedDict(Genes_Sorted)
    prev_stop = 0
    for pos, detail in Genes_Sorted.items():
        if detail[1] < prev_stop:
            Genes_Sorted[pos], Genes_Sorted[pos-1] = Genes_Sorted[pos-1], Genes_Sorted[pos]
        prev_stop = detail[1]
    return Genes_Sorted


def gff_load(options,gff_in,dna_regions):
    count = 0
    for line in gff_in:  # Get gene loci from GFF - ID=Gene will also classify Pseudogenes as genes
        line_data = line.split('\t')
        if line.startswith('\n') or line.startswith('#') or 'European Nucleotide Archive' in line:  # Not to crash on empty lines in GFF
            continue
        elif options.gene_ident[0] == 'ID=gene':
            if line_data[0] in dna_regions and options.gene_ident[0] in line_data[8]:
                start = int(line_data[3])
                stop = int(line_data[4])
                strand = line_data[6]
                gene_details = [start,stop,strand]
                dna_regions[line_data[0]][2].append({count:gene_details}) # This will add to list
                count += 1
        else:
            try:
                if line_data[2] == 'region':
                    continue
                elif line_data[0] in dna_regions:
                    if any(gene_type in line_data[2] for gene_type in options.gene_ident): # line[2] for normal run
                        start = int(line_data[3])
                        stop = int(line_data[4])
                        strand = line_data[6]
                        gene_details = [start, stop, strand]
                        if gene_details not in dna_regions[line_data[0]][2]:
                            dna_regions[line_data[0]][2].append({count:gene_details}) # This will add to list
                            count += 1
            except IndexError:
                continue
    return dna_regions


def fasta_load(fasta_in):
    dna_regions = collections.OrderedDict()
    first = True
    if '>' in fasta_in.readline().rstrip():
        fasta_in.seek(0)
        #### Default for when presented with standard fasta file
        for line in fasta_in:
            line = line.strip()
            if line.startswith('>') and first == False:  # Check if first seq in file
                dna_region_length = len(seq)
                dna_regions.update({dna_region_id: (seq, dna_region_length, list(), None)})
                seq = ''
                dna_region_id = line.split()[0].replace('>', '')
            elif line.startswith('>'):
                seq = ''
                dna_region_id = line.split()[0].replace('>', '')
            else:
                seq += str(line)
                first = False
        dna_region_length = len(seq)
        dna_regions.update({dna_region_id: (seq, dna_region_length, list(), None)})
    elif '##' in  fasta_in.readline().rstrip(): # Clunky and may fall over
        fasta_in.seek(0)
        #### Called when presented with Prokka GFF file so must get fasta from inside it
        ### Get to genome seq
        at_FASTA = False
        for line in fasta_in:  # Get gene loci from GFF - ID=Gene will also classify Pseudogenes as genes
            if line.startswith('##FASTA'):  # Not to crash on empty lines in GFF
                at_FASTA = True
            elif at_FASTA == True:
                line = line.strip()
                if line.startswith('>') and first == False:  # Check if first seq in file
                    dna_region_length = len(seq)
                    dna_regions.update({dna_region_id: (seq, dna_region_length, list(), None)})
                    seq = ''
                    dna_region_id = line.split()[0].replace('>', '')
                elif line.startswith('>'):
                    seq = ''
                    dna_region_id = line.split()[0].replace('>', '')
                else:
                    seq += str(line)
                    first = False
        dna_region_length = len(seq)
        dna_regions.update({dna_region_id: (seq, dna_region_length, list(), None)})

    return dna_regions


def get_rep_metrics(result):
    rep_metric_description = ('Percentage_of_Genes_Detected,Percentage_of_ORFs_that_Detected_a_Gene,'
                              'Percent_Difference_of_All_ORFs,Median_Length_Difference,Percentage_of_Perfect_Matches,'
                              'Median_Start_Difference_of_Matched_ORFs,Median_Stop_Difference_of_Matched_ORFs,'
                              'Percentage_Difference_of_Matched_Overlapping_CDSs,Percent_Difference_of_Short-Matched-ORFs,'
                              'Precision,Recall,False_Discovery_Rate')
    rep_metrics = [result['rep_metrics']['Percentage_of_Genes_Detected'],
                   result['rep_metrics']['Percentage_of_ORFs_that_Detected_a_Gene'],
                   result['rep_metrics']['Percent_Difference_of_All_ORFs'],
                   result['rep_metrics']['Median_Length_Difference'],
                   result['rep_metrics']['Percentage_of_Perfect_Matches'],
                   result['rep_metrics']['Median_Start_Difference_of_Matched_ORFs'],
                   result['rep_metrics']['Median_Stop_Difference_of_Matched_ORFs'],
                   result['rep_metrics']['Percentage_Difference_of_Matched_Overlapping_CDSs'],
                   result['rep_metrics']['Percent_Difference_of_Short-Matched-ORFs'],
                   result['rep_metrics']['Precision'],
                   result['rep_metrics']['Recall'],
                   result['rep_metrics']['False_Discovery_Rate']]
    return rep_metric_description, rep_metrics


def get_all_metrics(result):
    all_metric_description = ('Number_of_ORFs,Percent_Difference_of_All_ORFs,Number_of_ORFs_that_Detected_a_Gene,'
         'Percentage_of_ORFs_that_Detected_a_Gene,Number_of_Genes_Detected,Percentage_of_Genes_Detected,'
         'Median_Length_of_All_ORFs,Median_Length_Difference,Minimum_Length_of_All_ORFs,Minimum_Length_Difference,'
         'Maximum_Length_of_All_ORFs,Maximum_Length_Difference,Median_GC_content_of_All_ORFs,'
         'Percent_Difference_of_All_ORFs_Median_GC,Median_GC_content_of_Matched_ORFs,'
         'Percent_Difference_of_Matched_ORF_GC,Number_of_ORFs_which_Overlap_Another_ORF,'
         'Percent_Difference_of_Overlapping_ORFs,Maximum_ORF_Overlap,Median_ORF_Overlap,'
         'Number_of_Matched_ORFs_Overlapping_Another_ORF,Percentage_Difference_of_Matched_Overlapping_CDSs,'
         'Maximum_Matched_ORF_Overlap,Median_Matched_ORF_Overlap,Number_of_Short-ORFs,Percent_Difference_of_Short-ORFs,'
         'Number_of_Short-Matched-ORFs,Percent_Difference_of_Short-Matched-ORFs,Number_of_Perfect_Matches,'
         'Percentage_of_Perfect_Matches,Number_of_Perfect_Starts,Percentage_of_Perfect_Starts,Number_of_Perfect_Stops,'
         'Percentage_of_Perfect_Stops,Number_of_Out_of_Frame_ORFs,Number_of_Matched_ORFs_Extending_a_Coding_Region,'
         'Percentage_of_Matched_ORFs_Extending_a_Coding_Region,Number_of_Matched_ORFs_Extending_Start_Region,'
         'Percentage_of_Matched_ORFs_Extending_Start_Region,Number_of_Matched_ORFs_Extending_Stop_Region,'
         'Percentage_of_Matched_ORFs_Extending_Stop_Region,Number_of_All_ORFs_on_Positive_Strand,'
         'Percentage_of_All_ORFs_on_Positive_Strand,Number_of_All_ORFs_on_Negative_Strand,'
         'Percentage_of_All_ORFs_on_Negative_Strand,Median_Start_Difference_of_Matched_ORFs,'
         'Median_Stop_Difference_of_Matched_ORFs,ATG_Start_Percentage,GTG_Start_Percentage,TTG_Start_Percentage,'
         'ATT_Start_Percentage,CTG_Start_Percentage,Other_Start_Codon_Percentage,TAG_Stop_Percentage,'
         'TAA_Stop_Percentage,TGA_Stop_Percentage,Other_Stop_Codon_Percentage,True_Positive,False_Positive,'
         'False_Negative,Precision,Recall,False_Discovery_Rate,Nucleotide_True_Positive,Nucleotide_False_Positive,'
         'Nucleotide_True_Negative,Nucleotide_False_Negative,Nucleotide_Precision,Nucleotide_Recall,'
         'Nucleotide_False_Discovery_Rate,ORF_Nucleotide_Coverage_of_Genome,Matched_ORF_Nucleotide_Coverage_of_Genome')
    all_metrics = rep_metrics = [result['pred_metrics']['Number_of_ORFs'],
                                    result['pred_metrics']['Percent_Difference_of_All_ORFs'],
                                    result['pred_metrics']['Number_of_ORFs_that_Detected_a_Gene'],
                                    result['pred_metrics']['Percentage_of_ORFs_that_Detected_a_Gene'],
                                    result['pred_metrics']['Number_of_Genes_Detected'],
                                    result['pred_metrics']['Percentage_of_Genes_Detected'],
                                    result['pred_metrics']['Median_Length_of_All_ORFs'],
                                    result['pred_metrics']['Median_Length_Difference'],
                                    result['pred_metrics']['Minimum_Length_of_All_ORFs'],
                                    result['pred_metrics']['Minimum_Length_Difference'],
                                    result['pred_metrics']['Maximum_Length_of_All_ORFs'],
                                    result['pred_metrics']['Maximum_Length_Difference'],
                                    result['pred_metrics']['Median_GC_content_of_All_ORFs'],
                                    result['pred_metrics']['Percent_Difference_of_All_ORFs_Median_GC'],
                                    result['pred_metrics']['Median_GC_content_of_Matched_ORFs'],
                                    result['pred_metrics']['Percent_Difference_of_Matched_ORF_GC'],
                                    result['pred_metrics']['Number_of_ORFs_which_Overlap_Another_ORF'],
                                    result['pred_metrics']['Percent_Difference_of_Overlapping_ORFs'],
                                    result['pred_metrics']['Maximum_ORF_Overlap'],
                                    result['pred_metrics']['Median_ORF_Overlap'],
                                    result['pred_metrics']['Number_of_Matched_ORFs_Overlapping_Another_ORF'],
                                    result['pred_metrics']['Percentage_Difference_of_Matched_Overlapping_CDSs'],
                                    result['pred_metrics']['Maximum_Matched_ORF_Overlap'],
                                    result['pred_metrics']['Median_Matched_ORF_Overlap'],
                                    result['pred_metrics']['Number_of_Short-ORFs'],
                                    result['pred_metrics']['Percent_Difference_of_Short-ORFs'],
                                    result['pred_metrics']['Number_of_Short-Matched-ORFs'],
                                    result['pred_metrics']['Percent_Difference_of_Short-Matched-ORFs'],
                                    result['pred_metrics']['Number_of_Perfect_Matches'],
                                    result['pred_metrics']['Percentage_of_Perfect_Matches'],
                                    result['pred_metrics']['Number_of_Perfect_Starts'],
                                    result['pred_metrics']['Percentage_of_Perfect_Starts'],
                                    result['pred_metrics']['Number_of_Perfect_Stops'],
                                    result['pred_metrics']['Percentage_of_Perfect_Stops'],
                                    result['pred_metrics']['Number_of_Out_of_Frame_ORFs'],
                                    result['pred_metrics']['Number_of_Matched_ORFs_Extending_a_Coding_Region'],
                                    result['pred_metrics']['Percentage_of_Matched_ORFs_Extending_a_Coding_Region'],
                                    result['pred_metrics']['Number_of_Matched_ORFs_Extending_Start_Region'],
                                    result['pred_metrics']['Percentage_of_Matched_ORFs_Extending_Start_Region'],
                                    result['pred_metrics']['Number_of_Matched_ORFs_Extending_Stop_Region'],
                                    result['pred_metrics']['Percentage_of_Matched_ORFs_Extending_Stop_Region'],
                                    result['pred_metrics']['Number_of_All_ORFs_on_Positive_Strand'],
                                    result['pred_metrics']['Percentage_of_All_ORFs_on_Positive_Strand'],
                                    result['pred_metrics']['Number_of_All_ORFs_on_Negative_Strand'],
                                    result['pred_metrics']['Percentage_of_All_ORFs_on_Negative_Strand'],
                                    result['pred_metrics']['Median_Start_Difference_of_Matched_ORFs'],
                                    result['pred_metrics']['Median_Stop_Difference_of_Matched_ORFs'],
                                    result['pred_metrics']['ATG_Start_Percentage'],
                                    result['pred_metrics']['GTG_Start_Percentage'],
                                    result['pred_metrics']['TTG_Start_Percentage'],
                                    result['pred_metrics']['ATT_Start_Percentage'],
                                    result['pred_metrics']['CTG_Start_Percentage'],
                                    result['pred_metrics']['Other_Start_Codon_Percentage'],
                                    result['pred_metrics']['TAG_Stop_Percentage'],
                                    result['pred_metrics']['TAA_Stop_Percentage'],
                                    result['pred_metrics']['TGA_Stop_Percentage'],
                                    result['pred_metrics']['Other_Stop_Codon_Percentage'],
                                    result['pred_metrics']['True_Positive'],
                                    result['pred_metrics']['False_Positive'],
                                    result['pred_metrics']['False_Negative'],
                                    result['pred_metrics']['Precision'],
                                    result['pred_metrics']['Recall'],
                                    result['pred_metrics']['False_Discovery_Rate'],
                                    result['pred_metrics']['Nucleotide_True_Positive'],
                                    result['pred_metrics']['Nucleotide_False_Positive'],
                                    result['pred_metrics']['Nucleotide_True_Negative'],
                                    result['pred_metrics']['Nucleotide_False_Negative'],
                                    result['pred_metrics']['Nucleotide_Precision'],
                                    result['pred_metrics']['Nucleotide_Recall'],
                                    result['pred_metrics']['Nucleotide_False_Discovery_Rate'],
                                    result['pred_metrics']['ORF_Nucleotide_Coverage_of_Genome'],
                                    result['pred_metrics']['Matched_ORF_Nucleotide_Coverage_of_Genome']]


    return all_metric_description, all_metrics


from importlib import import_module
import  os

try:
    from .utils import *
except (ImportError, ModuleNotFoundError):
    from utils import *





def main():
    print('ORForise ' + ORForise_Version + ': List Tools Run Parameters')

    tools = set()
    base_dirs = [
        os.path.join(os.path.dirname(__file__), 'Tools'),
        os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'Tools')),
    ]

    for base in base_dirs:
        if not os.path.isdir(base):
            continue
        try:
            for entry in os.listdir(base):
                entry_path = os.path.join(base, entry)
                if os.path.isdir(entry_path) and not entry.startswith('.') and entry != '__pycache__':
                    tools.add(entry)
        except OSError:
            continue

    if not tools:
        print('No tools found in the searched directories.')
        return

    print('Available tools:')
    for tool_name in sorted(tools):
        print(' -', tool_name)
        try:
            tool_ = import_module('Tools.' + tool_name + '.' + tool_name)
            print('    Imported from Tools.' + tool_name)
        except ModuleNotFoundError:
            try:
                tool_ = import_module('ORForise.Tools.' + tool_name + '.' + tool_name)
                print('    Imported from ORForise.Tools.' + tool_name)
            except ModuleNotFoundError:
                print('    Tool not importable')



if __name__ == "__main__":
    try:
        try:
            main()
        except Exception:
            print('Unhandled exception in main')
    finally:
        print(CLOSING)


import collections

try:
    from utils import revCompIterative
    from utils import sortORFs
except ImportError:
    from ORForise.utils import revCompIterative
    from ORForise.utils import sortORFs


def Prokka(*args): # UNFINISHED
    tool_pred = args[0]
    dna_regions = args[1]
    types = args[2]
    if not dna_regions: # This triggers if dna_regions is an empty dict (GFF_Intersect passed nothing)
        dna_regions = collections.OrderedDict()
        with open(tool_pred, 'r') as PROKKA_input:
            for line in PROKKA_input:
                line = line.split()
                if "Prodigal" in line[1] and "CDS" in line[2] and line[0] not in dna_regions:
                    dna_regions[line[0]] = []  # Placeholder for genome sequence
        return dna_regions

    prokkaORFs = collections.defaultdict(list)
    for dna_region in dna_regions:
        prokkaORFs[dna_region] = collections.OrderedDict()
    for dna_region in dna_regions:
        try:
            genome = dna_regions[dna_region][0]
        except IndexError:
            genome = dna_regions[dna_region]
        genome_size = len(genome)
        genome_rev = revCompIterative(genome)
        with open(tool_pred, 'r') as prodigal_input:
            for line in prodigal_input:
                if '#' not in line:
                    line = line.split('\t')
                    if "prokka" not in line[1] and line[8].startswith('ID=') and dna_region in line[0] and "CDS" in line[2]:
                        start = int(line[3])
                        stop = int(line[4])
                        strand = line[6]
                        info = line[8]
                        if '-' in strand:  # Reverse Compliment starts and stops adjusted
                            r_start = genome_size - stop
                            r_stop = genome_size - start
                            startCodon = genome_rev[r_start:r_start + 3]
                            stopCodon = genome_rev[r_stop - 2:r_stop + 1]
                        elif '+' in strand:
                            startCodon = genome[start - 1:start + 2]
                            stopCodon = genome[stop - 3:stop]
                        po = str(start) + ',' + str(stop)
                        orf = [strand, startCodon, stopCodon, line[2], 'Prokka']
                        prokkaORFs.update({po: orf})

    for group in prokkaORFs:
        prokkaORFs[group] = sortORFs(prokkaORFs[group])
    return prokkaORFs
